from ...coordinates import CartesianCoordinates
from ...coordinates.origin import Origin
from ...time import Timestamp
from ...utils.helpers import orbits as orbits_helpers
from ..orbits import Orbits


def test_orbits__init__():
    coordinates = CartesianCoordinates.from_kwargs(
        x=[0.5],
        y=[0.5],
        z=[0.004],
        vx=[0.005],
        vy=[-0.005],
        vz=[0.0002],
        time=Timestamp.from_mjd([59000.0], scale="tdb"),
        origin=Origin.from_kwargs(code=["SUN"]),
        frame="ecliptic",
    )

    orbits = Orbits.from_kwargs(
        coordinates=coordinates, orbit_id=["1"], object_id=["Test Orbit"]
    )

    assert orbits.orbit_id[0].as_py() == "1"
    assert orbits.object_id[0].as_py() == "Test Orbit"
    assert orbits.coordinates.time.days.to_numpy()[0] == 59000
    assert orbits.coordinates.time.nanos.to_numpy()[0] == 0
    assert orbits.coordinates.time.scale == "tdb"
    assert orbits.coordinates.x.to_numpy()[0] == 0.5
    assert orbits.coordinates.y.to_numpy()[0] == 0.5
    assert orbits.coordinates.z.to_numpy()[0] == 0.004
    assert orbits.coordinates.vx.to_numpy()[0] == 0.005
    assert orbits.coordinates.vy.to_numpy()[0] == -0.005
    assert orbits.coordinates.vz.to_numpy()[0] == 0.0002


def test_orbit_iteration():
    orbits = orbits_helpers.make_simple_orbits(num_orbits=10)
    for o in orbits:
        assert len(o) == 1



import astropy.time
import astropy.units
import numpy as np
import numpy.testing as npt
import pyarrow as pa
import pyarrow.compute as pc
import pytest
import quivr as qv

from ..time import Timestamp


class Wrapper(qv.Table):
    id = qv.LargeStringColumn()
    times = Timestamp.as_column(nullable=True)


class TestTimeUnits:

    ts = Timestamp.from_kwargs(
        days=[0, 10000, 20000],
        nanos=[0, 1_000_000_000, 2_000_000_000],
    )

    empty = Timestamp.empty()

    with_nulls = qv.concatenate(
        [
            Wrapper.from_kwargs(times=ts, id=["a", "b", "c"]),
            Wrapper.from_kwargs(times=None, id=["d", "e", "f"]),
        ]
    ).times

    def test_micros(self):
        have = self.ts.micros()
        assert have.to_pylist() == [0, 1_000_000, 2_000_000]

        have = self.empty.micros()
        assert len(have) == 0

        have = self.with_nulls.micros()
        assert have.to_pylist() == [0, 1_000_000, 2_000_000, None, None, None]

    def test_millis(self):
        have = self.ts.millis()
        assert have.to_pylist() == [0, 1000, 2000]

        have = self.empty.millis()
        assert len(have) == 0

        have = self.with_nulls.millis()
        assert have.to_pylist() == [0, 1000, 2000, None, None, None]

    def test_seconds(self):
        have = self.ts.seconds()
        assert have.to_pylist() == [0, 1, 2]

        have = self.empty.seconds()
        assert len(have) == 0

        have = self.with_nulls.seconds()
        assert have.to_pylist() == [0, 1, 2, None, None, None]


class TestMJDConversions:
    ts = Timestamp.from_kwargs(
        days=[0, 10000, 20000],
        nanos=[0, 43200_000_000_000, 86400_000_000_000 - 1],
    )

    empty = Timestamp.empty()

    with_nulls = qv.concatenate(
        [
            Wrapper.from_kwargs(times=ts, id=["a", "b", "c"]),
            Wrapper.from_kwargs(times=None, id=["d", "e", "f"]),
        ]
    ).times

    def test_mjd(self):
        have = self.ts.mjd()
        assert have.to_pylist() == [0, 10000.5, 20000.999999999999]

    def test_empty(self):
        have = self.empty.mjd()
        assert len(have) == 0

    def test_with_nulls(self):
        have = self.with_nulls.mjd()
        assert have.to_pylist() == [0, 10000.5, 20000.999999999999, None, None, None]

    def test_from_list_ints(self):
        data = [0, 10000, 20000]
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 0, 0]

    def test_from_list_floats(self):
        data = [0.0, 10000.5, 20000.75]
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 43200_000_000_000, 64800_000_000_000]

    def test_from_numpy_ints(self):
        data = np.array([0, 10000, 20000], dtype=np.int64)
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 0, 0]

    def test_from_numpy_floats(self):
        data = np.array([0.0, 10000.5, 20000.75], dtype=np.float64)
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 43200_000_000_000, 64800_000_000_000]

    def test_from_pyarrow_ints(self):
        data = pa.array([0, 10000, 20000], type=pa.int64())
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 0, 0]

    def test_from_pyarrow_floats(self):
        data = pa.array([0.0, 10000.5, 20000.75], type=pa.float64())
        have = Timestamp.from_mjd(data)
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 43200_000_000_000, 64800_000_000_000]

    def test_from_pyarrow_table_float(self):
        """
        This test case exercises chunked arrays, which have slightly different behavior
        """
        data = pa.table({"mjd": [0.0, 10000.5, 20000.75]})
        have = Timestamp.from_mjd(data["mjd"])
        assert have.days.to_pylist() == [0, 10000, 20000]
        assert have.nanos.to_pylist() == [0, 43200_000_000_000, 64800_000_000_000]


class TestAstropyTime:
    ts = Timestamp.from_kwargs(
        days=[0, 50000, 60000],
        nanos=[0, 43200_000_000_000, 86400_000_000_000 - 1],
    )

    empty = Timestamp.empty()

    def test_to_astropy(self):
        have = self.ts.to_astropy()
        want = astropy.time.Time(
            [
                "1858-11-17T00:00:00.000000000",
                "1995-10-10T12:00:00.000000000",
                "2023-02-25T23:59:59.999999999",
            ],
            scale="tai",
        )

        npt.assert_allclose(have.mjd, want.mjd)

    def test_from_astropy(self):
        at = astropy.time.Time(
            [
                "1858-11-17T00:00:00.000000000",
                "1995-10-10T12:00:00.000000000",
                "2023-02-25T23:59:59.999999999",
            ],
            scale="tai",
        )
        have = Timestamp.from_astropy(at)
        assert have == self.ts

    def test_to_astropy_singleton(self):
        have = self.ts[0].to_astropy()
        want = astropy.time.Time("1858-11-17T00:00:00.000000000", scale="tai")
        assert have.mjd == want.mjd

    def test_from_astropy_scalar(self):
        at = astropy.time.Time("1858-11-17T00:00:00.000000000", scale="tai")
        have = Timestamp.from_astropy(at)
        assert have == self.ts[0]

    def test_empty(self):
        have = self.empty.to_astropy()
        assert len(have) == 0

    def test_roundtrip_scalar_zero(self):
        zero_astropy_time = astropy.time.Time(0, val2=0, format="mjd")

        from_at = Timestamp.from_astropy(zero_astropy_time)
        assert from_at.days[0].as_py() == 0
        assert from_at.nanos[0].as_py() == 0

        roundtrip = from_at.to_astropy()
        assert zero_astropy_time == roundtrip

    def test_roundtrip_second_precision(self):
        t1 = astropy.time.Time("2020-01-01T00:00:00.000000000", scale="tai")
        t2 = t1 + 1 * astropy.units.second

        have1 = Timestamp.from_astropy(t1)
        have2 = have1.add_seconds(1)

        assert have2.to_astropy() == t2


class TestTimeMath:

    t1 = Timestamp.from_kwargs(
        days=[0, 50000, 60000],
        nanos=[0, 43200_000_000_000, 86400_000_000_000 - 1],
    )

    def test_add_nanos_out_of_range(self):
        MIN_VAL = -86400_000_000_000
        MAX_VAL = 86400_000_000_000 - 1
        # Scalars:
        with pytest.raises(ValueError):
            self.t1.add_nanos(MIN_VAL - 1)
        with pytest.raises(ValueError):
            self.t1.add_nanos(MAX_VAL + 1)

        # Arrays:
        with pytest.raises(ValueError):
            self.t1.add_nanos([MIN_VAL - 1, 0, 0])
        with pytest.raises(ValueError):
            self.t1.add_nanos([0, MAX_VAL + 1, 0])
        with pytest.raises(ValueError):
            self.t1.add_nanos([0, 0, MIN_VAL - 1])
        with pytest.raises(ValueError):
            self.t1.add_nanos([0, 0, MAX_VAL + 1])

    def test_add_seconds_out_of_range(self):
        MIN_VAL = -86400
        MAX_VAL = 86400 - 1
        # Scalars:
        with pytest.raises(ValueError):
            self.t1.add_seconds(MIN_VAL - 1)
        with pytest.raises(ValueError):
            self.t1.add_seconds(MAX_VAL + 1)

        # Arrays:
        with pytest.raises(ValueError):
            self.t1.add_seconds([MIN_VAL - 1, 0, 0])
        with pytest.raises(ValueError):
            self.t1.add_seconds([0, MAX_VAL + 1, 0])

    def test_add_millis_out_of_range(self):
        MIN_VAL = -86400_000
        MAX_VAL = 86400_000 - 1
        # Scalars:
        with pytest.raises(ValueError):
            self.t1.add_millis(MIN_VAL - 1)
        with pytest.raises(ValueError):
            self.t1.add_millis(MAX_VAL + 1)

        # Arrays:
        with pytest.raises(ValueError):
            self.t1.add_millis([MIN_VAL - 1, 0, 0])
        with pytest.raises(ValueError):
            self.t1.add_millis([0, MAX_VAL + 1, 0])

    def test_add_micros_out_of_range(self):
        MIN_VAL = -86400_000_000
        MAX_VAL = 86400_000_000 - 1
        # Scalars:
        with pytest.raises(ValueError):
            self.t1.add_micros(MIN_VAL - 1)
        with pytest.raises(ValueError):
            self.t1.add_micros(MAX_VAL + 1)

        # Arrays:
        with pytest.raises(ValueError):
            self.t1.add_micros([MIN_VAL - 1, 0, 0])
        with pytest.raises(ValueError):
            self.t1.add_micros([0, MAX_VAL + 1, 0])

    def test_add_nanos_scalar(self):
        have = self.t1.add_nanos(1)
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [1, 43200_000_000_001, 0]

        have = self.t1.add_nanos(-1)
        assert have.days.to_pylist() == [-1, 50000, 60000]
        assert have.nanos.to_pylist() == [
            86400_000_000_000 - 1,
            43200_000_000_000 - 1,
            86400_000_000_000 - 2,
        ]

        have = self.t1.add_nanos(43200_000_000_000)
        assert have.days.to_pylist() == [0, 50001, 60001]
        assert have.nanos.to_pylist() == [43200_000_000_000, 0, 43200_000_000_000 - 1]

    def test_add_nanos_array(self):
        have = self.t1.add_nanos([1, 2, 3])
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [1, 43200_000_000_002, 2]

        have = self.t1.add_nanos([-1, -2, -3])
        assert have.days.to_pylist() == [-1, 50000, 60000]
        assert have.nanos.to_pylist() == [
            86400_000_000_000 - 1,
            43200_000_000_000 - 2,
            86400_000_000_000 - 4,
        ]

    def test_add_seconds(self):
        have = self.t1.add_seconds(1)
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [
            1_000_000_000,
            43200_000_000_000 + 1_000_000_000,
            1_000_000_000 - 1,
        ]

        have = self.t1.add_seconds(-1)
        assert have.days.to_pylist() == [-1, 50000, 60000]
        assert have.nanos.to_pylist() == [
            86400_000_000_000 - 1_000_000_000,
            43200_000_000_000 - 1_000_000_000,
            86400_000_000_000 - 1_000_000_000 - 1,
        ]

        have = self.t1.add_seconds([1, 2, 3])
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [
            1_000_000_000,
            43200_000_000_000 + 2_000_000_000,
            3_000_000_000 - 1,
        ]

    def test_add_millis(self):
        have = self.t1.add_millis(1)
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [
            1_000_000,
            43200_000_000_000 + 1_000_000,
            1_000_000 - 1,
        ]

        have = self.t1.add_millis(-1)
        assert have.days.to_pylist() == [-1, 50000, 60000]
        assert have.nanos.to_pylist() == [
            86400_000_000_000 - 1_000_000,
            43200_000_000_000 - 1_000_000,
            86400_000_000_000 - 1_000_000 - 1,
        ]

        have = self.t1.add_millis([1, 2, 3])
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [
            1_000_000,
            43200_000_000_000 + 2_000_000,
            3_000_000 - 1,
        ]

    def test_add_micros(self):
        have = self.t1.add_micros(1)
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [1_000, 43200_000_000_000 + 1_000, 1_000 - 1]

        have = self.t1.add_micros(-1)
        assert have.days.to_pylist() == [-1, 50000, 60000]
        assert have.nanos.to_pylist() == [
            86400_000_000_000 - 1_000,
            43200_000_000_000 - 1_000,
            86400_000_000_000 - 1_000 - 1,
        ]

        have = self.t1.add_micros([1, 2, 3])
        assert have.days.to_pylist() == [0, 50000, 60001]
        assert have.nanos.to_pylist() == [1_000, 43200_000_000_000 + 2_000, 3_000 - 1]

    def test_add_days(self):
        have = self.t1.add_days(1)
        assert have.days.to_pylist() == [1, 50001, 60001]
        assert have.nanos.to_pylist() == self.t1.nanos.to_pylist()

        have = self.t1.add_days(-1)
        assert have.days.to_pylist() == [-1, 49999, 59999]
        assert have.nanos.to_pylist() == self.t1.nanos.to_pylist()

        have = self.t1.add_days([1, 2, 3])
        assert have.days.to_pylist() == [1, 50002, 60003]
        assert have.nanos.to_pylist() == self.t1.nanos.to_pylist()

    def test_equals_array(self):
        t1 = self.t1
        t2 = self.t1

        assert pc.all(t1.equals_array(t2)).as_py()

        t3 = t2.add_nanos(1)
        assert not pc.all(t1.equals_array(t3)).as_py()
        assert pc.all(t1.equals_array(t3, precision="us")).as_py()

        t4 = Timestamp.from_kwargs(
            days=t1.days,
            nanos=t1.nanos,
            scale="utc",
        )
        with pytest.raises(ValueError):
            t1.equals_array(t4)

    def test_equals_scalar(self):
        t1 = Timestamp.from_kwargs(
            days=[50000, 60000, 70000],
            nanos=[0, 1, 2],
        )

        have = t1.equals_scalar(days=50000, nanos=0)
        assert have.to_pylist() == [True, False, False]

    def test_equals_scalar_precision(self):
        t1 = Timestamp.from_kwargs(
            days=[0, 0, 1, 1, 2, 2],
            nanos=[
                500,
                86400_000_000_000 - 500,
                500,
                86400_000_000_000 - 500,
                500,
                86400_000_000_000 - 500,
            ],
        )
        have = t1.equals_scalar(days=1, nanos=0, precision="us")
        assert have.to_pylist() == [False, True, True, False, False, False]

    def test_difference_scalar(self):
        # Compute difference from days=1, nanos=100
        cases = [
            {
                "in_days": 0,
                "in_nanos": 0,
                "out_days": -2,
                "out_nanos": 86400_000_000_000 - 100,
            },
            {
                "in_days": 0,
                "in_nanos": 50,
                "out_days": -2,
                "out_nanos": 86400_000_000_000 - 50,
            },
            {"in_days": 0, "in_nanos": 200, "out_days": -1, "out_nanos": 100},
            {
                "in_days": 1,
                "in_nanos": 50,
                "out_days": -1,
                "out_nanos": 86400_000_000_000 - 50,
            },
            {"in_days": 1, "in_nanos": 100, "out_days": 0, "out_nanos": 0},
            {"in_days": 1, "in_nanos": 200, "out_days": 0, "out_nanos": 100},
            {
                "in_days": 2,
                "in_nanos": 50,
                "out_days": 0,
                "out_nanos": 86400_000_000_000 - 50,
            },
            {"in_days": 2, "in_nanos": 100, "out_days": 1, "out_nanos": 0},
            {"in_days": 2, "in_nanos": 200, "out_days": 1, "out_nanos": 100},
        ]

        for i, c in enumerate(cases):
            t1 = Timestamp.from_kwargs(
                days=[c["in_days"]],
                nanos=[c["in_nanos"]],
            )
            have_days, have_nanos = t1.difference_scalar(days=1, nanos=100)
            assert have_days[0].as_py() == c["out_days"], f"case {i}"
            assert have_nanos[0].as_py() == c["out_nanos"], f"case {i}"

    def test_difference(self):
        t1 = Timestamp.from_kwargs(
            days=[50000, 60000, 70000],
            nanos=[0, 1, 2],
        )
        t2 = Timestamp.from_kwargs(
            days=[50000, 60000, 70000],
            nanos=[100, 200, 300],
        )
        have_days, have_nanos = t1.difference(t2)
        assert have_days.to_pylist() == [-1, -1, -1]
        assert have_nanos.to_pylist() == [
            86400_000_000_000 - 100,
            86400_000_000_000 - 199,
            86400_000_000_000 - 298,
        ]


def test_dataframe_roundtrip():
    t1 = Timestamp.from_kwargs(
        days=[50000, 60000, 70000],
        nanos=[0, 1, 2],
        scale="tdb",
    )

    df = t1.to_dataframe()

    t2 = Timestamp.from_dataframe(df)

    assert t1 == t2
    assert t1.scale == t2.scale


def test_dataframe_roundtrip_nested():
    t1 = Timestamp.from_kwargs(
        days=[50000, 60000, 70000],
        nanos=[0, 1, 2],
        scale="tdb",
    )
    w = Wrapper.from_kwargs(
        id=["a", "b", "c"],
        times=t1,
    )

    df = w.to_dataframe()

    w2 = Wrapper.from_flat_dataframe(df)

    assert w == w2
    assert w.times.scale == w2.times.scale


class TestUnique:
    def test_empty(self):
        t = Timestamp.empty(scale="tdb")
        have = t.unique()
        assert have.scale == "tdb"
        assert len(have) == 0

        t2 = Timestamp.empty(scale="utc")
        have = t2.unique()
        assert have.scale == "utc"
        assert len(have) == 0

    def test_one(self):
        t = Timestamp.from_kwargs(days=[1], nanos=[2], scale="tdb")
        have = t.unique()
        assert have.scale == "tdb"
        assert len(have) == 1
        assert have.days.to_pylist() == [1]
        assert have.nanos.to_pylist() == [2]

    def test_multple(self):
        t = Timestamp.from_kwargs(
            days=[1, 1, 2, 2, 1, 1], nanos=[2, 3, 4, 5, 2, 3], scale="tdb"
        )
        have = t.unique()
        assert have.scale == "tdb"
        assert len(have) == 4
        pairs = zip(have.days.to_pylist(), have.nanos.to_pylist())
        assert sorted(list(pairs)) == [(1, 2), (1, 3), (2, 4), (2, 5)]


@pytest.mark.parametrize("rescale", [True, False])
def test_Timestamp_link(rescale):
    # Test that the link method works as expected
    time = Timestamp.from_kwargs(
        days=[68000, 68000, 68010, 68010, 68020, 68020],
        nanos=[1, 2, 3, 3, 4, 4],
        scale="tdb",
    )

    other = Timestamp.from_kwargs(
        days=[68000, 68000, 68010, 68020], nanos=[1, 2, 3, 4], scale="tdb"
    )
    if rescale:
        other = other.rescale("utc")

    linkage = time.link(other)
    assert len(linkage.all_unique_values) == 4

    key = (68000, 1)
    left_table, right_table = linkage.select(key)
    assert len(left_table) == 1
    assert len(right_table) == 1
    assert left_table.days.to_pylist() == [68000]
    assert left_table.nanos.to_pylist() == [1]
    assert right_table.days.to_pylist() == [68000]
    assert right_table.nanos.to_pylist() == [1]

    key = (68000, 2)
    left_table, right_table = linkage.select(key)
    assert len(left_table) == 1
    assert len(right_table) == 1
    assert left_table.days.to_pylist() == [68000]
    assert left_table.nanos.to_pylist() == [2]
    assert right_table.days.to_pylist() == [68000]
    assert right_table.nanos.to_pylist() == [2]

    key = (68010, 3)
    left_table, right_table = linkage.select(key)
    assert len(left_table) == 2
    assert len(right_table) == 1
    assert left_table.days.to_pylist() == [68010, 68010]
    assert left_table.nanos.to_pylist() == [3, 3]
    assert right_table.days.to_pylist() == [68010]
    assert right_table.nanos.to_pylist() == [3]

    key = (68020, 4)
    left_table, right_table = linkage.select(key)
    assert len(left_table) == 2
    assert len(right_table) == 1
    assert left_table.days.to_pylist() == [68020, 68020]
    assert left_table.nanos.to_pylist() == [4, 4]
    assert right_table.days.to_pylist() == [68020]
    assert right_table.nanos.to_pylist() == [4]


def test_Timestamp_link_precision():
    # Test that the link method works as expected with user defined precisions
    time = Timestamp.from_kwargs(
        days=[1, 1, 1, 1, 1, 1, 1, 1],
        nanos=[9, 99, 999, 9_999, 99_999, 999_999, 9_999_999, 99_999_999],
        scale="tdb",
    )

    # 1 nano per nano :)
    time_ns = Timestamp.from_kwargs(
        days=[1, 1, 1, 1, 1, 1, 1, 1],
        nanos=[9, 99, 999, 9_999, 99_999, 999_999, 9_999_999, 99_999_999],
        scale="tdb",
    )
    linkage = time.link(time_ns, precision="ns")
    assert len(linkage.all_unique_values) == 8

    # 1_000 nanos per micro
    time_micros = Timestamp.from_kwargs(
        days=[1, 1, 1, 1, 1, 1, 1, 1],
        nanos=[0, 0, 0, 9_999, 99_999, 999_999, 9_999_999, 99_999_999],
        scale="tdb",
    )
    linkage = time.link(time_micros, precision="us")
    assert len(linkage.all_unique_values) == 6

    # 1_000_000 nanos per milli
    time_millis = Timestamp.from_kwargs(
        days=[1, 1, 1, 1, 1, 1, 1, 1],
        nanos=[0, 0, 0, 0, 0, 0, 9_999_999, 99_999_999],
        scale="tdb",
    )
    linkage = time.link(time_millis, precision="ms")
    assert len(linkage.all_unique_values) == 3

    # 1_000_000_000 nanos per second
    time_s = Timestamp.from_kwargs(
        days=[1, 1, 1, 1, 1, 1, 1, 1],
        nanos=[
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        scale="tdb",
    )
    linkage = time.link(time_s, precision="s")
    assert len(linkage.all_unique_values) == 1


def test_Timestamp_max():
    # Test that the max method correctly returns
    # the maximum time
    times = Timestamp.from_kwargs(
        days=[5, 1, 3, 2, 4, 1, 2, 3, 4, 5],
        nanos=[2, 2, 2, 2, 2, 1, 1, 1, 1, 1],
    )
    times = qv.concatenate([times for _ in range(10)])

    max_time = times.max()
    assert len(max_time) == 1
    assert max_time.scale == times.scale
    assert max_time.days.to_pylist() == [5]
    assert max_time.nanos.to_pylist() == [2]


def test_Timestamp_min():
    # Test that the min method correctly returns
    # the minimum time
    times = Timestamp.from_kwargs(
        days=[5, 1, 3, 2, 4, 1, 2, 3, 4, 5],
        nanos=[2, 2, 2, 2, 2, 1, 1, 1, 1, 1],
    )
    times = qv.concatenate([times for _ in range(10)])

    min_time = times.min()
    assert len(min_time) == 1
    assert min_time.scale == times.scale
    assert min_time.days.to_pylist() == [1]
    assert min_time.nanos.to_pylist() == [1]



import numpy as np
import pyarrow.compute as pc
import pytest
import quivr as qv
from astropy import units as u

from ...observers import Observers
from ...time import Timestamp
from ..ephemeris import generate_ephemeris_2body

OBJECT_IDS = [
    "594913 'Aylo'chaxnim (2020 AV2)",
    "163693 Atira (2003 CP20)",
    "(2010 TK7)",
    "3753 Cruithne (1986 TO)",
    "54509 YORP (2000 PH5)",
    "2063 Bacchus (1977 HB)",
    "1221 Amor (1932 EA1)",
    "433 Eros (A898 PA)",
    "3908 Nyx (1980 PA)",
    "434 Hungaria (A898 RB)",
    "1876 Napolitania (1970 BA)",
    "2001 Einstein (1973 EB)",
    "2 Pallas (A802 FA)",
    "6 Hebe (A847 NA)",
    "6522 Aci (1991 NQ)",
    "10297 Lynnejones (1988 RJ13)",
    "17032 Edlu (1999 FM9)",
    "202930 Ivezic (1998 SG172)",
    "911 Agamemnon (A919 FB)",
    "1143 Odysseus (1930 BH)",
    "1172 Aneas (1930 UA)",
    "3317 Paris (1984 KF)",
    "5145 Pholus (1992 AD)",
    "5335 Damocles (1991 DA)",
    "15760 Albion (1992 QB1)",
    "15788 (1993 SB)",
    "15789 (1993 SC)",
    "1I/'Oumuamua (A/2017 U1)",
]

TOLERANCES = {
    # range (m), angular difference (mas), light_time difference (microseconds)
    # Default range is about consistent with our ability to get
    # observatory state vector positions compared to Horizons
    "default": (20, 0.1, 10),
    # 594913 'Aylo'chaxnim (2020 AV2) : Vatira
    # Range difference for first 30 observations between
    # 32.8 m and 80.4 m, for last 60 observations range is
    # between 0.13 m and 26.3 m
    "594913 'Aylo'chaxnim (2020 AV2)": (100, 1, 10),
    # 3753 Cruithne (1986 TO) : NEO (ATEN)
    # Range anywhere between 0.01 m and 40.5 m throughout
    # the 90 observations
    "3753 Cruithne (1986 TO)": (50, 1, 10),
    # 6522 Aci (1991 NQ) : Inner Main Belt asteroid
    # 89/90 RA, Decs have a mean offset of 0.007 mas
    # 1/90 RA, Decs has an offset of 1.24 mas
    # 88/90 ranges have a mean offset of 10.6 m
    # 2/90 ranges are 89.5, 174.3 m
    "6522 Aci (1991 NQ)": (200, 2, 10),
    # 1143 Odysseus (1930 BH) : Jupiter Trojan
    # 89/90 light times have a mean offset of 0.14 microseconds
    # 1/90 light times has an offset of 38.4 microseconds
    # 89/90 RA, Decs have a mean offset of 0.0056 mas
    # 1/90 RA, Decs has an offset of 8.9 mas
    # 89/90 ranges have a mean offset of 6.9 m
    # 1/90 ranges has an offset of 11575 m? ...
    "1143 Odysseus (1930 BH)": (11600, 10, 50),
}


@pytest.mark.parametrize("object_id", OBJECT_IDS)
def test_generate_ephemeris_2body(object_id, propagated_orbits, ephemeris):
    # For our catalog of test orbits, generate ephemeris using Horizons generated state vectors
    # and compare the results to the Horizons generated ephemeris
    propagated_orbit = propagated_orbits.select("object_id", object_id)
    ephemeris_orbit = ephemeris[ephemeris["targetname"].isin([object_id])]

    observers_list = []
    for observatory_code in ephemeris_orbit["observatory_code"].unique():
        observatory_mask = ephemeris_orbit["observatory_code"] == observatory_code
        observer_i = Observers.from_code(
            observatory_code,
            Timestamp.from_mjd(
                ephemeris_orbit[observatory_mask]["mjd_utc"].values,
                scale="utc",
            ),
        )
        observers_list.append(observer_i)

    observers = qv.concatenate(observers_list)

    ephemeris_orbit_2body = generate_ephemeris_2body(propagated_orbit, observers)

    # Get the tolerances for this orbit
    if object_id in TOLERANCES:
        range_tolerance, angular_tolerance, light_time_tolerance = TOLERANCES[object_id]
    else:
        range_tolerance, angular_tolerance, light_time_tolerance = TOLERANCES["default"]

    # Calculate the offset in light travel time
    light_time_diff = np.abs(
        ephemeris_orbit_2body.light_time.to_numpy()
        - (ephemeris_orbit["lighttime"].values * 60 / 86400)
    ) * (u.d).to(u.microsecond)

    # Assert that the light time is less than the tolerance
    np.testing.assert_array_less(light_time_diff, light_time_tolerance)

    # Calculate the difference in RA and Dec
    angular_diff = np.abs(
        ephemeris_orbit_2body.coordinates.values[:, 1:3]
        - ephemeris_orbit[["RA", "DEC"]].values
    ) * (u.deg).to(u.milliarcsecond)

    # Topocentric difference
    range_diff = np.abs(
        ephemeris_orbit_2body.coordinates.values[:, 0] - ephemeris_orbit["delta"].values
    ) * (u.au).to(u.m)

    # Assert that the position is less than the tolerance
    np.testing.assert_array_less(angular_diff, angular_tolerance)

    # Assert that the range is less than the tolerance
    np.testing.assert_array_less(range_diff, range_tolerance)

    # Ephemeris 2-body does not propagate the covariance or provide the
    # aberrated state so the aberrated coordinates should be empty
    assert ephemeris_orbit_2body.aberrated_coordinates.frame == "unspecified"
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.x)).as_py()
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.y)).as_py()
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.z)).as_py()
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.vx)).as_py()
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.vy)).as_py()
    assert pc.all(pc.is_null(ephemeris_orbit_2body.aberrated_coordinates.vz)).as_py()



import uuid

import quivr as qv

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.residuals import Residuals
from ..orbits.orbits import Orbits


class FittedOrbits(qv.Table):

    orbit_id = qv.LargeStringColumn(default=lambda: uuid.uuid4().hex)
    object_id = qv.LargeStringColumn(nullable=True)
    coordinates = CartesianCoordinates.as_column()
    arc_length = qv.Float64Column()
    num_obs = qv.Int64Column()
    chi2 = qv.Float64Column()
    reduced_chi2 = qv.Float64Column()
    iterations = qv.Int64Column(nullable=True)
    success = qv.BooleanColumn(nullable=True)
    status_code = qv.Int64Column(nullable=True)

    def to_orbits(self) -> Orbits:
        """
        Convert fitted orbits to orbits that can be used by
        a Propagator.

        Returns
        -------
        orbits : `~adam_core.orbits.Orbits`
            Orbits.
        """
        return Orbits.from_kwargs(
            orbit_id=self.orbit_id,
            object_id=self.object_id,
            coordinates=self.coordinates,
        )


class FittedOrbitMembers(qv.Table):

    orbit_id = qv.LargeStringColumn()
    obs_id = qv.LargeStringColumn()
    residuals = Residuals.as_column(nullable=True)
    solution = qv.BooleanColumn(nullable=True)
    outlier = qv.BooleanColumn(nullable=True)



import warnings
from typing import Union

import pandas as pd
import quivr as qv
from mpc_obscodes import mpc_obscodes
from typing_extensions import Self

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.origin import OriginCodes
from ..time import Timestamp


class ObservatoryGeodetics(qv.Table):
    code = qv.LargeStringColumn()
    longitude = qv.Float64Column()
    cos_phi = qv.Float64Column()
    sin_phi = qv.Float64Column()
    name = qv.LargeStringColumn()


# Read MPC extended observatory codes file
# Ignore warning about pandas deprecation
with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        message="The behavior of 'to_datetime' with 'unit' when parsing strings is deprecated",
    )
    OBSCODES = pd.read_json(
        mpc_obscodes,
        orient="index",
        dtype={"Longitude": float, "cos": float, "sin": float, "Name": str},
        encoding_errors="strict",
        precise_float=True,
    )
    OBSCODES.reset_index(inplace=True, names=["code"])

OBSERVATORY_GEODETICS = ObservatoryGeodetics.from_kwargs(
    code=OBSCODES["code"].values,
    longitude=OBSCODES["Longitude"].values,
    cos_phi=OBSCODES["cos"].values,
    sin_phi=OBSCODES["sin"].values,
    name=OBSCODES["Name"].values,
)

OBSERVATORY_CODES = {
    x for x in OBSERVATORY_GEODETICS.code.to_numpy(zero_copy_only=False)
}


class Observers(qv.Table):
    code = qv.LargeStringColumn(nullable=False)
    coordinates = CartesianCoordinates.as_column()

    @classmethod
    def from_code(cls, code: Union[str, OriginCodes], times: Timestamp) -> Self:
        """
        Instantiate an Observers table with a single code and multiple times.
        Times do not need to be unique. The observer state will be calculated
        for each time and correctly matched to the input times and replicated for
        duplicate times.

        To load multiple codes, use `from_code` and then concatenate the tables.

        Note that NAIF origin codes may not be supported by `~adam_core.propagator.Propagator`
        classes such as PYOORB.

        Parameters
        ----------
        code : Union[str, OriginCodes]
            MPC observatory code or NAIF origin code for which to find the states.
        times : Timetamp (N)
            Epochs for which to find the observatory locations.

        Returns
        -------
        observers : `~adam_core.observers.observers.Observers` (N)
            The observer and its state at each time.

        Examples
        --------
        >>> import numpy as np
        >>> from adam_core.time import Timestamp
        >>> from adam_core.observers import Observers
        >>> times = Timestamp.from_mjd(np.arange(59000, 59000 + 100), scale="tdb")
        >>> observers = Observers.from_code("X05", times)
        """
        from .state import get_observer_state

        if isinstance(code, OriginCodes):
            code_str = code.name
        elif isinstance(code, str):
            code_str = code
        else:
            err = "Code should be a str or an `~adam_core.coordinates.origin.OriginCodes`."
            raise ValueError(err)

        return cls.from_kwargs(
            code=[code_str] * len(times),
            coordinates=get_observer_state(code, times),
        )

    def iterate_codes(self):
        """
        Iterate over the codes in the Observers table.

        Yields
        ------
        code : str
            The code for observer.
        observers : `~adam_core.observers.observers.Observers`
            The Observers table for this observer.
        """
        for code in self.code.unique().sort():
            yield code.as_py(), self.select("code", code)



import os

import pytest

try:
    from adam_core.propagator.adam_pyoorb import PYOORBPropagator
except ImportError:
    PYOORBPropagator = None

from ..differential_correction import fit_least_squares


@pytest.mark.skipif(
    os.environ.get("OORB_DATA") is None, reason="OORB_DATA environment variable not set"
)
@pytest.mark.skipif(PYOORBPropagator is None, reason="PYOORBPropagator not available")
def test_fit_least_squares_pure_iod_orbit(pure_iod_orbit):
    # Test that fit_least_squares can fit and improve a pure orbit from an IOD
    # process using least squares

    orbit, orbit_members, observations = pure_iod_orbit
    propagator = PYOORBPropagator()

    fitted_orbit, fitted_orbit_members = fit_least_squares(
        orbit, observations, propagator
    )

    assert len(fitted_orbit) == 1
    assert len(fitted_orbit_members) == len(orbit_members) == len(observations)
    assert fitted_orbit.reduced_chi2[0].as_py() < (orbit.reduced_chi2[0].as_py() / 1e4)
    assert fitted_orbit.status_code[0].as_py() > 0
    assert fitted_orbit.iterations[0].as_py() <= 50



import numpy as np
import numpy.testing as npt
import pytest

from ..tisserand import calc_tisserand_parameter

# --- Tests last updated: 2023-02-09

# From JPL's SBDB (2023-02-01) https://ssd.jpl.nasa.gov/sbdb.cgi
# a, e, i
DAMOCLES = {
    "aei": (11.87052898293103, 0.8659079806381217, 61.6840796497527),
    "T_jupiter": 1.155,
}
CERES = {
    "aei": (2.767181743149466, 0.07881745101960996, 10.58634326912728),
    "T_jupiter": 3.310,
}
# 3D/Biela
BIELA = {
    "aei": (3.53465808340135, 0.751299, 13.2164),
    "T_jupiter": 2.531,
}


def test_calc_tisserand_parameter_damocloids():
    Tp = calc_tisserand_parameter(*DAMOCLES["aei"], third_body="jupiter")

    npt.assert_allclose(np.round(Tp, 3), DAMOCLES["T_jupiter"])
    return


def test_calc_tisserand_parameter_asteroids():
    Tp = calc_tisserand_parameter(*CERES["aei"], third_body="jupiter")

    npt.assert_allclose(np.round(Tp, 3), CERES["T_jupiter"])
    return


def test_calc_tisserand_parameter_jupiter_family_comets():
    Tp = calc_tisserand_parameter(*BIELA["aei"], third_body="jupiter")

    npt.assert_allclose(np.round(Tp, 3), BIELA["T_jupiter"])
    return


def test_calc_tisserand_parameter_raise():
    # Not a valid planet name
    with pytest.raises(ValueError):
        calc_tisserand_parameter(*CERES["aei"], third_body="")



# flake8: noqa: F401

# TODO: calc_orbit_class does not work currently, but would be nice as a method on Orbits
from .ephemeris import Ephemeris
from .orbits import Orbits
from .variants import VariantOrbits

__all__ = [
    "Ephemeris",
    "Orbits",
    "VariantOrbits",
]



import pyarrow as pa
import pytest
import quivr as qv

from ...coordinates.origin import Origin
from ...coordinates.spherical import SphericalCoordinates
from ...observers.observers import Observers
from ...time import Timestamp
from ..ephemeris import Ephemeris


def test_Ephemeris_link_to_observers():
    # Test that we can link ephemerides to observers with different
    # precisions
    observer_01 = Observers.from_code(
        "X05",
        Timestamp.from_kwargs(
            days=[59000, 59000, 59000],
            nanos=[9, 999, 999_999],
            scale="utc",
        ),
    )
    observer_02 = Observers.from_code(
        "500",
        Timestamp.from_kwargs(
            days=[59000, 59000, 59000],
            nanos=[9, 999, 999_999],
            scale="utc",
        ),
    )
    observers = qv.concatenate([observer_01, observer_02])

    ephemeris = Ephemeris.from_kwargs(
        orbit_id=pa.array(["00000" for i in range(6)]),
        object_id=pa.array(["00000" for i in range(6)]),
        coordinates=SphericalCoordinates.from_kwargs(
            time=Timestamp.from_kwargs(
                days=[59000, 59000, 59000, 59000, 59000, 59000],
                nanos=[9, 999, 999_999, 9, 999, 999_999],
                scale="utc",
            ),
            lon=[0, 0, 0, 0, 0, 0],
            lat=[0, 0, 0, 0, 0, 0],
            frame="equatorial",
            origin=Origin.from_kwargs(
                code=pa.array(["X05", "X05", "X05", "500", "500", "500"])
            ),
        ),
    )

    linkage = ephemeris.link_to_observers(observers, precision="ns")
    assert len(linkage.all_unique_values) == 6
    e1, o1 = linkage.select((59000, 9, "X05"))
    assert len(e1) == len(o1) == 1
    e2, o2 = linkage.select((59000, 999, "X05"))
    assert len(e2) == len(o2) == 1
    e3, o3 = linkage.select((59000, 999_999, "X05"))
    assert len(e3) == len(o3) == 1
    e4, o4 = linkage.select((59000, 9, "500"))
    assert len(e4) == len(o4) == 1
    e5, o5 = linkage.select((59000, 999, "500"))
    assert len(e5) == len(o5) == 1
    e6, o6 = linkage.select((59000, 999_999, "500"))
    assert len(e6) == len(o6) == 1

    # Reduce precision to microseconds
    with pytest.warns(UserWarning):
        linkage = ephemeris.link_to_observers(observers, precision="us")

    # First two times should be grouped together
    # Last 2 times should be round-down to the previous microsecond
    assert len(linkage.all_unique_values) == 4
    e1, o1 = linkage.select((59000, 0, "X05"))
    assert len(e1) == len(o1) == 2
    e2, o2 = linkage.select((59000, 0, "500"))
    assert len(e2) == len(o2) == 2
    e3, o3 = linkage.select((59000, 999_000, "X05"))
    assert len(e3) == len(o3) == 1
    e4, o4 = linkage.select((59000, 999_000, "500"))
    assert len(e4) == len(o4) == 1

    # Reduce precision to milliseconds
    with pytest.warns(UserWarning):
        linkage = ephemeris.link_to_observers(observers, precision="ms")

    assert len(linkage.all_unique_values) == 2
    e1, o1 = linkage.select((59000, 0, "X05"))
    assert len(e1) == len(o1) == 3
    e2, o2 = linkage.select((59000, 0, "500"))
    assert len(e2) == len(o2) == 3



import numpy as np
import numpy.testing as npt
import pytest

from ...time import Timestamp
from ..keplerian import KeplerianCoordinates
from ..origin import Origin


@pytest.fixture
def keplerian_elements(orbital_elements):
    return KeplerianCoordinates.from_kwargs(
        a=orbital_elements["a"],
        e=orbital_elements["e"],
        i=orbital_elements["incl"],
        raan=orbital_elements["Omega"],
        ap=orbital_elements["w"],
        M=orbital_elements["M"],
        time=Timestamp.from_mjd(orbital_elements["mjd_tdb"].values, scale="tdb"),
        origin=Origin.from_kwargs(code=["SUN"] * len(orbital_elements)),
        frame="ecliptic",
    )


def test_KeplerianCoordinates_period(keplerian_elements, orbital_elements):
    # Test that the period calculated from Keplerian elements matches the
    # period in the test data.
    P_desired = orbital_elements["P"]
    # For parabolic/hyperbolic orbits, orbital period is infinite
    # Our test data from JPL Horizons uses a value of 1e99 for infinite
    P_desired = np.where(P_desired > 1e99, np.inf, P_desired)
    P_actual = keplerian_elements.P
    npt.assert_allclose(P_actual, P_desired, rtol=0.0, atol=1e-10)


def test_KeplerianCoordinates_period_setter(keplerian_elements):
    # Assert that we cannot set the period
    with pytest.raises(ValueError, match="Cannot set period (P)*"):
        keplerian_elements.P = np.full(len(keplerian_elements), np.nan)


def test_KeplerianCoordinates_period_deleter(keplerian_elements):
    # Assert that we cannot delete the period
    with pytest.raises(ValueError, match="Cannot delete period (P)*"):
        del keplerian_elements.P


def test_KeplerianCoordinates_perapsis_distance(keplerian_elements, orbital_elements):
    # Test that the periapsis distance calculated from Keplerian elements
    # matches the periapsis distance in the test data.
    q_desired = orbital_elements["q"]
    q_actual = keplerian_elements.q
    npt.assert_allclose(q_actual, q_desired, rtol=0.0, atol=1e-12)


def test_KeplerianCoordinates_perapsis_distance_setter(keplerian_elements):
    # Assert that we cannot set the periapsis distance
    with pytest.raises(ValueError, match="Cannot set periapsis distance (q)*"):
        keplerian_elements.q = np.full(len(keplerian_elements), np.nan)


def test_KeplerianCoordinates_perapsis_distance_deleter(keplerian_elements):
    # Assert that we cannot delete the periapsis distance
    with pytest.raises(ValueError, match="Cannot delete periapsis distance (q)*"):
        del keplerian_elements.q


def test_KeplerianCoordinates_apoapsis_distance(keplerian_elements, orbital_elements):
    # Test that the apoapsis distance calculated from Keplerian elements
    # matches the apoapsis distance in the test data.
    Q_desired = orbital_elements["Q"]
    # For parabolic/hyperbolic orbits, apoapsis distance is infinite
    # Our test data from JPL Horizons uses a value of 1e99 for infinite
    Q_desired = np.where(Q_desired > 1e99, np.inf, Q_desired)
    Q_actual = keplerian_elements.Q
    npt.assert_allclose(Q_actual, Q_desired, rtol=0.0, atol=1e-12)


def test_KeplerianCoordinates_apoapsis_distance_setter(keplerian_elements):
    # Assert that we cannot set the apoapsis distance
    with pytest.raises(ValueError, match="Cannot set apoapsis distance (Q)*"):
        keplerian_elements.Q = np.full(len(keplerian_elements), np.nan)


def test_KeplerianCoordinates_apoapsis_distance_deleter(keplerian_elements):
    # Assert that we cannot delete the apoapsis distance
    with pytest.raises(ValueError, match="Cannot delete apoapsis distance (Q)*"):
        del keplerian_elements.Q


def test_KeplerianCoordinates_semi_latus_rectum(keplerian_elements, orbital_elements):
    # Test that the semi-latus rectum calculated from Keplerian elements
    # matches the semi-latus rectum in the test data.
    a = orbital_elements["a"].values
    e = orbital_elements["e"].values
    p_desired = (1 - e**2) * a
    p_actual = keplerian_elements.p
    npt.assert_allclose(p_actual, p_desired, rtol=0.0, atol=1e-12)


def test_KeplerianCoordinates_semi_latus_rectum_setter(keplerian_elements):
    # Assert that we cannot set the semi-latus rectum
    with pytest.raises(ValueError, match="Cannot set semi-latus rectum (p)*"):
        keplerian_elements.p = np.full(len(keplerian_elements), np.nan)


def test_KeplerianCoordinates_semi_latus_rectum_deleter(keplerian_elements):
    # Assert that we cannot delete the semi-latus rectum
    with pytest.raises(ValueError, match="Cannot delete semi-latus rectum (p)*"):
        del keplerian_elements.p


def test_KeplerianCoordinates_mean_motion(keplerian_elements, orbital_elements):
    # Test that the mean motion calculated from Keplerian elements
    # matches the mean motion in the test data.
    n_desired = orbital_elements["n"]
    n_actual = keplerian_elements.n
    npt.assert_allclose(n_actual, n_desired, rtol=0.0, atol=1e-12)


def test_KeplerianCoordinates_mean_motion_setter(keplerian_elements):
    # Assert that we cannot set the mean motion
    with pytest.raises(ValueError, match="Cannot set mean motion (n)*"):
        keplerian_elements.n = np.full(len(keplerian_elements), np.nan)


def test_KeplerianCoordinates_mean_motion_deleter(keplerian_elements):
    # Assert that we cannot delete the mean motion
    with pytest.raises(ValueError, match="Cannot delete mean motion (n)*"):
        del keplerian_elements.n



import numpy as np
import spiceypy as sp
from astropy import units as u

from ..origin import Origin
from ..transform import cartesian_to_keplerian, keplerian_to_cartesian


def test_keplerian_to_cartesian_elliptical_against_spice(orbital_elements):
    # Test keplerian_to_cartesian (vmapped) against the cartesian elements
    # calculated by SPICE for a series of sample orbital elements.

    # Limit to elliptical orbits
    orbital_elements = orbital_elements[orbital_elements["e"] < 1]

    keplerian_elements = orbital_elements[["a", "e", "incl", "Omega", "w", "M"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    cartesian_elements_actual = keplerian_to_cartesian(keplerian_elements, origin.mu())

    # Keplerian elements for SPICE are expected to be defined with periapse distance
    # not the semi-major axis
    t0_jd = t0 + 2400000.5
    keplerian_elements_spice = np.empty((len(keplerian_elements), 8))
    keplerian_elements_spice[:, 0] = orbital_elements["q"].values
    keplerian_elements_spice[:, 1] = orbital_elements["e"].values
    keplerian_elements_spice[:, 2] = np.radians(orbital_elements["incl"].values)
    keplerian_elements_spice[:, 3] = np.radians(orbital_elements["Omega"].values)
    keplerian_elements_spice[:, 4] = np.radians(orbital_elements["w"].values)
    keplerian_elements_spice[:, 5] = np.radians(orbital_elements["M"].values)
    keplerian_elements_spice[:, 6] = t0_jd
    keplerian_elements_spice[:, 7] = origin.mu()

    cartesian_elements_spice = np.empty((len(keplerian_elements), 6))
    for i in range(len(keplerian_elements)):
        cartesian_elements_spice[i] = sp.conics(keplerian_elements_spice[i], t0_jd[i])

    diff = cartesian_elements_actual - cartesian_elements_spice

    # Calculate offset in position in mm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.mm)
    # Calculate offset in velocity in nm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.nm / u.s)

    # Assert positions are to within 20 mm
    np.testing.assert_array_less(r_diff, 20)
    # Assert velocities are to within 10 nm/s
    np.testing.assert_array_less(v_diff, 10)


def test_keplerian_to_cartesian_hyperbolic_against_spice(orbital_elements):
    # Test keplerian_to_cartesian (vmapped) against the cartesian elements
    # calculated by SPICE for a series of sample orbital elements.

    # Limit to hyperbolic orbits
    orbital_elements = orbital_elements[orbital_elements["e"] > 1]

    keplerian_elements = orbital_elements[["a", "e", "incl", "Omega", "w", "M"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    cartesian_elements_actual = keplerian_to_cartesian(keplerian_elements, origin.mu())

    # Keplerian elements for SPICE are expected to be defined with periapse distance
    # not the semi-major axis
    t0_jd = t0 + 2400000.5
    keplerian_elements_spice = np.empty((len(keplerian_elements), 8))
    keplerian_elements_spice[:, 0] = orbital_elements["q"].values
    keplerian_elements_spice[:, 1] = orbital_elements["e"].values
    keplerian_elements_spice[:, 2] = np.radians(orbital_elements["incl"].values)
    keplerian_elements_spice[:, 3] = np.radians(orbital_elements["Omega"].values)
    keplerian_elements_spice[:, 4] = np.radians(orbital_elements["w"].values)
    keplerian_elements_spice[:, 5] = np.radians(orbital_elements["M"].values)
    keplerian_elements_spice[:, 6] = t0_jd
    keplerian_elements_spice[:, 7] = origin.mu()

    cartesian_elements_spice = np.empty((len(keplerian_elements), 6))
    for i in range(len(keplerian_elements)):
        cartesian_elements_spice[i] = sp.conics(keplerian_elements_spice[i], t0_jd[i])

    diff = cartesian_elements_actual - cartesian_elements_spice

    # Calculate offset in position in mm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.mm)
    # Calculate offset in velocity in nm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.nm / u.s)

    # Assert positions are to within 10 mm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 10 nm/s
    np.testing.assert_array_less(v_diff, 10)


def test_cartesian_to_keplerian_elliptical_against_spice(orbital_elements):
    # Test cartesian_to_keplerian (vmapped) against the keplerian elements
    # calculated by SPICE for a series of sample orbital elements.

    # Limit to elliptical orbits
    orbital_elements = orbital_elements[orbital_elements["e"] < 1]

    t0 = orbital_elements["mjd_tdb"].values
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0_jd = t0 + 2400000.5
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))
    mu = origin.mu()

    # Cartesian to keplerian returns an N, 13 array containing the keplerian elements and cometary
    # elements, with some additional columns.
    keplerian_elements_actual = cartesian_to_keplerian(cartesian_elements, t0, mu)

    # SPICE returns periapse distance instead of semi-major axis so lets test
    # that as well
    q_actual = keplerian_elements_actual[:, 2]
    keplerian_elements_actual = keplerian_elements_actual[:, [0, 4, 5, 6, 7, 8]]

    keplerian_elements_spice = np.empty((len(cartesian_elements), 8))
    for i in range(len(keplerian_elements_spice)):
        keplerian_elements_spice[i] = sp.oscelt(
            np.copy(cartesian_elements[i]), t0_jd[i], mu[i]
        )

    # Keplerian elements from SPICE report periapse distance instead of semi-major axis
    # and angles are in radians instead of degrees, also returned are the gravitational
    # parameter and epoch (so lets ignore those)
    keplerian_elements_spice = keplerian_elements_spice[:, :6]
    q = keplerian_elements_spice[:, 0].copy()
    e = keplerian_elements_spice[:, 1]
    a = q / (1 - e)
    keplerian_elements_spice[:, 0] = a
    keplerian_elements_spice[:, 2:6] = np.degrees(keplerian_elements_spice[:, 2:6])

    # Calculate the difference
    diff = keplerian_elements_actual - keplerian_elements_spice

    # Calculate offset in semi-major axis in mm
    a_diff = np.abs(diff[:, 0]) * u.au.to(u.mm)
    # Assert semi-major axis is to within 10 mm
    np.testing.assert_array_less(a_diff, 10)

    # Calculate the offset in periapse distance in mm
    q_diff = np.abs(q_actual - q) * u.au.to(u.mm)
    # Assert periapse distance is to within 10 mm
    np.testing.assert_array_less(q_diff, 10)

    # Calculate offset in eccentricity
    e_diff = np.abs(diff[:, 1])
    # Assert eccentricity is to within 1e-15
    np.testing.assert_array_less(e_diff, 1e-15)

    # Calculate offset in inclination in nanoarcseconds
    i_diff = np.abs(diff[:, 2]) * u.degree.to(u.nanoarcsecond)
    # Assert inclination is to within 10 nanoarcseconds
    np.testing.assert_array_less(i_diff, 10)

    # Calculate offset in longitude of ascending node in nanoarcseconds
    raan_diff = np.abs(diff[:, 3]) * u.degree.to(u.nanoarcsecond)
    # Assert longitude of ascending node is to within 10 nanoarcseconds
    np.testing.assert_array_less(raan_diff, 10)

    # Calculate offset in argument of periapsis in nanoarcseconds
    ap_diff = np.abs(diff[:, 4]) * u.degree.to(u.nanoarcsecond)
    # Assert argument of periapsis is to within 10 nanoarcseconds
    np.testing.assert_array_less(ap_diff, 10)

    # Calculate offset in mean anomaly in nanoarcseconds
    M_diff = np.abs(diff[:, 5]) * u.degree.to(u.nanoarcsecond)
    # Assert mean anomaly is to within 10 nanoarcseconds
    np.testing.assert_array_less(M_diff, 10)


def test_cartesian_to_keplerian_hyperbolic_against_spice(orbital_elements):
    # Test cartesian_to_keplerian (vmapped) against the keplerian elements
    # calculated by SPICE for a series of sample orbital elements.

    # Limit to hyperbolic orbits
    orbital_elements = orbital_elements[orbital_elements["e"] > 1]

    t0 = orbital_elements["mjd_tdb"].values
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0_jd = t0 + 2400000.5
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))
    mu = origin.mu()

    # Cartesian to keplerian returns an N, 13 array containing the keplerian elements and cometary
    # elements, with some additional columns.
    keplerian_elements_actual = cartesian_to_keplerian(cartesian_elements, t0, mu)

    # SPICE returns periapse distance instead of semi-major axis so lets test
    # that as well
    q_actual = keplerian_elements_actual[:, 2]
    keplerian_elements_actual = keplerian_elements_actual[:, [0, 4, 5, 6, 7, 8]]

    keplerian_elements_spice = np.empty((len(cartesian_elements), 8))
    for i in range(len(keplerian_elements_spice)):
        keplerian_elements_spice[i] = sp.oscelt(
            np.copy(cartesian_elements[i]), t0_jd[i], mu[i]
        )

    # Keplerian elements from SPICE report periapse distance instead of semi-major axis
    # and angles are in radians instead of degrees, also returned are the gravitational
    # parameter and epoch (so lets ignore those)
    keplerian_elements_spice = keplerian_elements_spice[:, :6]
    q = keplerian_elements_spice[:, 0].copy()
    e = keplerian_elements_spice[:, 1]
    a = q / (1 - e)
    keplerian_elements_spice[:, 0] = a
    keplerian_elements_spice[:, 2:6] = np.degrees(keplerian_elements_spice[:, 2:6])

    # Calculate the difference
    diff = keplerian_elements_actual - keplerian_elements_spice

    # Calculate offset in semi-major axis in mm
    a_diff = np.abs(diff[:, 0]) * u.au.to(u.mm)
    # Assert semi-major axis is to within 10 mm
    np.testing.assert_array_less(a_diff, 10)

    # Calculate the offset in periapse distance in mm
    q_diff = np.abs(q_actual - q) * u.au.to(u.mm)
    # Assert periapse distance is to within 10 mm
    np.testing.assert_array_less(q_diff, 10)

    # Calculate offset in eccentricity
    e_diff = np.abs(diff[:, 1])
    # Assert eccentricity is to within 1e-15
    np.testing.assert_array_less(e_diff, 1e-15)

    # Calculate offset in inclination in nanoarcseconds
    i_diff = np.abs(diff[:, 2]) * u.degree.to(u.nanoarcsecond)
    # Assert inclination is to within 10 nanoarcseconds
    np.testing.assert_array_less(i_diff, 10)

    # Calculate offset in longitude of ascending node in nanoarcseconds
    raan_diff = np.abs(diff[:, 3]) * u.degree.to(u.nanoarcsecond)
    # Assert longitude of ascending node is to within 10 nanoarcseconds
    np.testing.assert_array_less(raan_diff, 10)

    # Calculate offset in argument of periapsis in nanoarcseconds
    ap_diff = np.abs(diff[:, 4]) * u.degree.to(u.nanoarcsecond)
    # Assert argument of periapsis is to within 10 nanoarcseconds
    np.testing.assert_array_less(ap_diff, 10)

    # Calculate offset in mean anomaly in nanoarcseconds
    M_diff = np.abs(diff[:, 5]) * u.degree.to(u.nanoarcsecond)
    # Assert mean anomaly is to within 10 nanoarcseconds
    np.testing.assert_array_less(M_diff, 10)



import numpy as np

from ..cartesian import CartesianCoordinates
from ..covariances import CoordinateCovariances
from ..origin import Origin


def test_CartesianCoordinates_rotate():
    # Test rotation of Cartesian coordinates. Rotating coordinates and their
    # covariances with an identity matrix should return the same coordinates and covariances.
    N, D = 1000, 6
    values = np.random.random((N, D))
    covariances = np.random.random((N, D, D))

    coords = CartesianCoordinates.from_kwargs(
        x=values[:, 0],
        y=values[:, 1],
        z=values[:, 2],
        vx=values[:, 3],
        vy=values[:, 4],
        vz=values[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances),
        origin=Origin.from_kwargs(code=["origin"] * N),
        frame="equatorial",
    )

    rot_matrix = np.identity(D)
    coords_rotated = coords.rotate(rot_matrix, "identity")

    np.testing.assert_equal(coords_rotated.values, values)
    np.testing.assert_equal(coords_rotated.covariance.to_matrix(), covariances)

    # Now repeat the same with a non-identity matrix (rotation by 90 degrees
    # about the x-axis)
    theta = np.radians(90)
    rot_matrix = np.array(
        [
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, np.cos(theta), -np.sin(theta), 0.0, 0.0, 0.0],
            [0.0, np.sin(theta), np.cos(theta), 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, np.cos(theta), -np.sin(theta)],
            [0.0, 0.0, 0.0, 0.0, np.sin(theta), np.cos(theta)],
        ]
    )

    coords_rotated = coords.rotate(rot_matrix, "x-axis-90")

    # x position and velocity should be unchanged
    np.testing.assert_equal(coords_rotated.values[:, 0], values[:, 0])
    np.testing.assert_equal(coords_rotated.values[:, 3], values[:, 3])

    # y rotates to z, z rotates to -y, vy rotates to -vz, vz rotates to vy
    np.testing.assert_almost_equal(coords_rotated.values[:, 2], values[:, 1])
    np.testing.assert_almost_equal(coords_rotated.values[:, 1], -values[:, 2])
    np.testing.assert_almost_equal(coords_rotated.values[:, 4], -values[:, 5])
    np.testing.assert_almost_equal(coords_rotated.values[:, 5], values[:, 4])

    # Covariances should also be rotated (this is a bit harder to test) but we
    # write out the explicit matrix multiplication)
    covariances_rotated = rot_matrix @ covariances @ rot_matrix.T
    np.testing.assert_almost_equal(
        coords_rotated.covariance.to_matrix(), covariances_rotated
    )
    return


def test_CartesianCoordinates_translate():
    # Test translation of Cartesian coordinates. Translating coordinates and their
    # covariances with a zero vector should return the same coordinates and covariances.
    N, D = 1000, 6
    values = np.random.random((N, D))
    covariances = np.random.random((N, D, D))

    coords = CartesianCoordinates.from_kwargs(
        x=values[:, 0],
        y=values[:, 1],
        z=values[:, 2],
        vx=values[:, 3],
        vy=values[:, 4],
        vz=values[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances),
        origin=Origin.from_kwargs(code=["origin"] * N),
        frame="equatorial",
    )

    translation_vector = np.zeros(D)
    coords_translated = coords.translate(translation_vector, "zero")

    np.testing.assert_equal(coords_translated.values, values)
    np.testing.assert_equal(coords_translated.covariance.to_matrix(), covariances)

    # Now repeat the same with a non-zero vector
    translation_vector = np.array([1.0, -2.0, 3.0, 4.0, -5.0, 6.0])
    coords_translated = coords.translate(translation_vector, "nonzero")

    np.testing.assert_equal(coords_translated.values, values + translation_vector)
    np.testing.assert_equal(coords_translated.covariance.to_matrix(), covariances)


def test_CartesianCoordinates_attributes():
    # Test that the correct position attributes are returned both for individual
    # dimensions but also multiple dimensions (as vectors)
    N, D = 1000, 6
    values = np.random.random((N, D))
    covariances = np.random.random((N, D, D))

    coords = CartesianCoordinates.from_kwargs(
        x=values[:, 0],
        y=values[:, 1],
        z=values[:, 2],
        vx=values[:, 3],
        vy=values[:, 4],
        vz=values[:, 5],
        covariance=CoordinateCovariances.from_matrix(covariances),
        origin=Origin.from_kwargs(code=["origin"] * N),
        frame="equatorial",
    )

    # Test coordinate dimensions
    np.testing.assert_equal(coords.x, values[:, 0])
    np.testing.assert_equal(coords.y, values[:, 1])
    np.testing.assert_equal(coords.z, values[:, 2])
    np.testing.assert_equal(coords.vx, values[:, 3])
    np.testing.assert_equal(coords.vy, values[:, 4])
    np.testing.assert_equal(coords.vz, values[:, 5])

    # Test coordinate vectors and vector magnitudes
    r_mag = np.linalg.norm(values[:, :3], axis=1)
    r_hat = values[:, :3] / r_mag[:, None]
    v_mag = np.linalg.norm(values[:, 3:], axis=1)
    v_hat = values[:, 3:] / v_mag[:, None]
    np.testing.assert_equal(coords.r, values[:, :3])
    np.testing.assert_equal(coords.r_mag, r_mag)
    np.testing.assert_equal(coords.r_hat, r_hat)
    np.testing.assert_equal(coords.v, values[:, 3:])
    np.testing.assert_equal(coords.v_mag, v_mag)
    np.testing.assert_equal(coords.v_hat, v_hat)

    # Test covariance sigmas
    np.testing.assert_equal(coords.sigma_x, np.sqrt(covariances[:, 0, 0]))
    np.testing.assert_equal(coords.sigma_y, np.sqrt(covariances[:, 1, 1]))
    np.testing.assert_equal(coords.sigma_z, np.sqrt(covariances[:, 2, 2]))
    np.testing.assert_equal(coords.sigma_vx, np.sqrt(covariances[:, 3, 3]))
    np.testing.assert_equal(coords.sigma_vy, np.sqrt(covariances[:, 4, 4]))
    np.testing.assert_equal(coords.sigma_vz, np.sqrt(covariances[:, 5, 5]))

    # Test position and velocity sigma vectors
    sigma_r = np.empty((N, 3))
    sigma_r[:, 0] = np.sqrt(covariances[:, 0, 0])
    sigma_r[:, 1] = np.sqrt(covariances[:, 1, 1])
    sigma_r[:, 2] = np.sqrt(covariances[:, 2, 2])
    np.testing.assert_equal(coords.sigma_r, sigma_r)

    sigma_r_mag = np.linalg.norm(sigma_r, axis=1)
    np.testing.assert_equal(coords.sigma_r_mag, sigma_r_mag)

    sigma_v = np.empty((N, 3))
    sigma_v[:, 0] = np.sqrt(covariances[:, 3, 3])
    sigma_v[:, 1] = np.sqrt(covariances[:, 4, 4])
    sigma_v[:, 2] = np.sqrt(covariances[:, 5, 5])
    np.testing.assert_equal(coords.sigma_v, sigma_v)

    sigma_v_mag = np.linalg.norm(sigma_v, axis=1)
    np.testing.assert_equal(coords.sigma_v_mag, sigma_v_mag)



import numpy.typing as npt
from astropy.time import Time

BASE62 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
BASE62_MAP = {BASE62[i]: i for i in range(len(BASE62))}


def _unpack_mpc_date(epoch_pf: str) -> Time:
    # Taken from Lynne Jones' SSO TOOLS.
    # See https://minorplanetcenter.net/iau/info/PackedDates.html
    # for MPC documentation on packed dates.
    # Examples:
    #    1998 Jan. 18.73     = J981I73
    #    2001 Oct. 22.138303 = K01AM138303
    epoch_pf = str(epoch_pf)
    year = int(epoch_pf[0], base=32) * 100 + int(epoch_pf[1:3])
    month = int(epoch_pf[3], base=32)
    day = int(epoch_pf[4], base=32)
    isot_string = "{:d}-{:02d}-{:02d}".format(year, month, day)

    if len(epoch_pf) > 5:
        fractional_day = float("." + epoch_pf[5:])
        hours = int((24 * fractional_day))
        minutes = int(60 * ((24 * fractional_day) - hours))
        seconds = 3600 * (24 * fractional_day - hours - minutes / 60)
        isot_string += "T{:02d}:{:02d}:{:09.6f}".format(hours, minutes, seconds)

    return Time(isot_string, format="isot", scale="tt")


def convert_mpc_packed_dates(pf_tt: npt.ArrayLike) -> Time:
    """
    Convert MPC packed form dates (in the TT time scale) to
    MJDs in TT. See: https://minorplanetcenter.net/iau/info/PackedDates.html
    for details on the packed date format.

    Parameters
    ----------
    pf_tt : `~numpy.ndarray` (N)
        MPC-style packed form epochs in the TT time scale.

    Returns
    -------
    mjd_tt : `~astropy.time.core.Time` (N)
        Epochs in TT MJDs.
    """
    isot_tt = []
    for epoch in pf_tt:
        isot_tt.append(_unpack_mpc_date(epoch))

    return Time(isot_tt)


def pack_numbered_designation(designation: str) -> str:
    """
    Pack a numbered MPC designation.

    Examples of numbered designations:
        Numbered      Packed
        3202          03202
        50000         50000
        100345        A0345
        360017        a0017
        203289        K3289
        620000        ~0000
        620061        ~000z
        3140113       ~AZaz
        15396335      ~zzzz

    Parameters
    ----------
    designation : str
        MPC numbered designation.

    Returns
    -------
    designation_pf : str
        MPC packed numbered designation.

    Raises
    ------
    ValueError : If the numbered designation cannot be packed.
        If the numbered designation is larger than 15396335.
    """
    number = int(designation)
    if number > 15396335:
        raise ValueError(
            "Numbered designation is too large. Maximum supported is 15396335."
        )

    if number <= 99999:
        return "{:05}".format(number)
    elif (number >= 100000) and (number <= 619999):
        bigpart, remainder = divmod(number, 10000)
        return f"{BASE62[bigpart]}{remainder:04}"
    else:
        x = number - 620000
        number_pf = []
        while x:
            number_pf.append(BASE62[int(x % 62)])
            x //= 62

        number_pf.reverse()
        return "~{}".format("".join(number_pf).zfill(4))


def pack_provisional_designation(designation: str) -> str:
    """
    Pack a provisional MPC designation.

    Examples of provisional designations:
        Provisional   Packed
        1995 XA       J95X00A
        1995 XL1      J95X01L
        1995 FB13     J95F13B
        1998 SQ108    J98SA8Q
        1998 SV127    J98SC7V
        1998 SS162    J98SG2S
        2099 AZ193    K99AJ3Z
        2008 AA360    K08Aa0A
        2007 TA418    K07Tf8A

    Parameters
    ----------
    designation : str
        MPC provisional designation.

    Returns
    -------
    designation_pf : str
        MPC packed provisional designation.

    Raises
    ------
    ValueError : If the provisional designation cannot be packed.
        The provisional designations is not at least 6 characters long.
        The first 4 characters of the provisional designation are not a year.
        The 5th character of the provisional designation is not a space.
        The provisional designation contains a hyphen.
        The half-month letter is I or Z.
    """
    if len(designation) < 6:
        raise ValueError(
            "Provisional designations should be at least 6 characters long."
        )
    if not designation[:3].isdecimal():
        raise ValueError(
            "Expected the first 4 characters of the provisional designation to be a year."
        )
    if designation[4] != " ":
        raise ValueError(
            "Expected the 5th character of the provisional designation to be a space."
        )
    if "-" in designation:
        raise ValueError("Provisional designations cannot contain a hyphen.")

    year = BASE62[int(designation[0:2])] + designation[2:4]
    letter1 = designation[5]
    letter2 = designation[6]
    cycle = designation[7:]

    if letter1 in {"I", "Z"}:
        raise ValueError("Half-month letters cannot be I or Z.")
    if letter1.isdecimal() or letter2.isdecimal():
        raise ValueError("Invalid provisional designation.")

    cycle_pf = "00"
    if len(cycle) > 0:
        cycle_int = int(cycle)
        if cycle_int <= 99:
            cycle_pf = str(cycle_int).zfill(2)
        else:
            cycle_pf = BASE62[cycle_int // 10] + str(cycle_int % 10)

    designation_pf = "{}{}{}{}".format(year, letter1, cycle_pf, letter2)
    return designation_pf


def pack_survey_designation(designation: str) -> str:
    """
    Pack a survey MPC designation.

    Examples of survey designations:
        Survey       Packed
        2040 P-L     PLS2040
        3138 T-1     T1S3138
        1010 T-2     T2S1010
        4101 T-3     T3S4101

    Parameters
    ----------
    designation : str
        MPC survey designation.

    Returns
    -------
    designation_pf : str
        MPC packed survey designation.

    Raises
    ------
    ValueError : If the survey designation cannot be packed.
        The survey designation does not start with P-L, T-1, T-2, or T-3.
    """
    number = designation[0:4]
    survey = designation[5:]

    if survey == "P-L":
        survey_pf = "PLS"

    elif survey[0:2] == "T-" and survey[2] in {"1", "2", "3"}:
        survey_pf = "T{}S".format(survey[2])

    else:
        raise ValueError("Survey designations must start with P-L, T-1, T-2, T-3.")

    designation_pf = "{}{}".format(survey_pf, number.zfill(4))
    return designation_pf


def pack_mpc_designation(designation: str) -> str:
    """
    Pack a unpacked MPC designation. For example, provisional
    designation 1998 SS162 will be packed to J98SG2S. Permanent
    designation 323 will be packed to 00323.

    TODO: add support for comet and natural satellite designations

    Parameters
    ----------
    designation : str
        MPC unpacked designation.

    Returns
    -------
    designation_pf : str
        MPC packed form designation.

    Raises
    ------
    ValueError : If designation cannot be packed.
    """
    # Lets see if its a numbered object
    try:
        return pack_numbered_designation(designation)
    except ValueError:
        pass

    # If its not numbered, maybe its a provisional designation
    try:
        return pack_provisional_designation(designation)
    except ValueError:
        pass

    # If its a survey designation, deal with it
    try:
        return pack_survey_designation(designation)
    except ValueError:
        pass

    err = (
        "Unpacked designation '{}' could not be packed.\n"
        "It could not be recognized as any of the following:\n"
        " - a numbered object (e.g. '3202', '203289', '3140113')\n"
        " - a provisional designation (e.g. '1998 SV127', '2008 AA360')\n"
        " - a survey designation (e.g. '2040 P-L', '3138 T-1')"
    )
    raise ValueError(err.format(designation))


def unpack_numbered_designation(designation_pf: str) -> str:
    """
    Unpack a numbered MPC designation.

    Examples of numbered designations:
        Numbered      Unpacked
        03202         3202
        50000         50000
        A0345         100345
        a0017         360017
        K3289         203289
        ~0000         620000
        ~000z         620061
        ~AZaz         3140113
        ~zzzz         15396335

    Parameters
    ----------
    designation_pf : str
        MPC packed numbered designation.

    Returns
    -------
    designation : str
        MPC unpacked numbered designation.

    Raises
    ------
    ValueError : If the numbered designation cannot be unpacked.
        The packed numbered designation is not at least 4 characters long.
    """
    number = None
    # Numbered objects (1 - 99999)
    if designation_pf.isdecimal():
        number = int(designation_pf)

    # Numbered objects (620000+)
    elif designation_pf[0] == "~":
        number = 620000
        number_pf = designation_pf[1:]
        for i, c in enumerate(number_pf):
            power = len(number_pf) - (i + 1)
            number += BASE62_MAP[c] * (62**power)

    # Numbered objects (100000 - 619999)
    else:
        number = BASE62_MAP[designation_pf[0]] * 10000 + int(designation_pf[1:])

    if number is None:
        raise ValueError("Packed numbered designation could not be unpacked.")
    else:
        designation = str(number)

    return designation


def unpack_provisional_designation(designation_pf: str) -> str:
    """
    Unpack a provisional MPC designation.

    Examples of provisional designations:
        Provisional   Unpacked
        J95X00A       1995 XA
        J95X01L       1995 XL1
        J95F13B       1995 FB13
        J98SA8Q       1998 SQ108
        J98SC7V       1998 SV127
        J98SG2S       1998 SS162
        K99AJ3Z       2099 AZ193
        K08Aa0A       2008 AA360
        K07Tf8A       2007 TA418

    Parameters
    ----------
    designation_pf : str
        MPC packed provisional designation.

    Returns
    -------
    designation : str
        MPC unpacked provisional designation.

    Raises
    ------
    ValueError : If the provisional designation cannot be unpacked.
        The packed provisional designation is not 7 characters long.
        The packed provisional designation does not have a year.
    """
    if len(designation_pf) != 7:
        raise ValueError("Provisional designation must be 7 characters long.")
    if not designation_pf[1].isdecimal() or not designation_pf[2].isdecimal():
        raise ValueError("Provisional designation must have a year.")
    year = str(BASE62_MAP[designation_pf[0]] * 100 + int(designation_pf[1:3]))
    letter1 = designation_pf[3]
    letter2 = designation_pf[6]
    if letter1.isdecimal() or letter2.isdecimal():
        raise ValueError()
    cycle1 = designation_pf[4]
    cycle2 = designation_pf[5]

    number = int(BASE62_MAP[cycle1]) * 10 + BASE62_MAP[cycle2]
    if number == 0:
        number_str = ""
    else:
        number_str = str(number)

    designation = "{} {}{}{}".format(year, letter1, letter2, number_str)

    return designation


def unpack_survey_designation(designation_pf: str) -> str:
    """
    Unpack a survey MPC designation.

    Examples of survey designations:
        Survey       Packed
        PLS2040      2040 P-L
        T1S3138      3138 T-1
        T2S1010      1010 T-2
        T3S4101      4101 T-3

    Parameters
    ----------
    designation_pf : str
        MPC packed survey designation.

    Returns
    -------
    designation : str
        MPC unpacked survey designation.

    Raises
    ------
    ValueError : If the survey designation cannot be unpacked.
        The packed survey designation does not start with PLS, T1S, T2S, or T3S.
    """
    number = int(designation_pf[3:8])
    survey_pf = designation_pf[0:3]
    if survey_pf not in {"PLS", "T1S", "T2S", "T3S"}:
        raise ValueError(
            "Packed survey designation must start with PLS, T1S, T2S, or T3S."
        )

    if survey_pf == "PLS":
        survey = "P-L"

    if survey_pf[0] == "T" and survey_pf[2] == "S":
        survey = "T-{}".format(survey_pf[1])

    designation = "{} {}".format(number, survey)
    return designation


def unpack_mpc_designation(designation_pf: str) -> str:
    """
    Unpack a packed MPC designation. For example, provisional
    designation J98SG2S will be unpacked to 1998 SS162. Permanent
    designation 00323 will be unpacked to 323.

    TODO: add support for comet and natural satellite designations

    Parameters
    ----------
    designation_pf : str
        MPC packed form designation.

    Returns
    -------
    designation : str
        MPC unpacked designation.

    Raises
    ------
    ValueError : If designation_pf cannot be unpacked.
    """
    # Lets see if its a numbered object
    try:
        return unpack_numbered_designation(designation_pf)
    except ValueError:
        pass

    # Lets see if its a provisional designation
    try:
        return unpack_provisional_designation(designation_pf)
    except ValueError:
        pass

    # Lets see if its a survey designation
    try:
        return unpack_survey_designation(designation_pf)
    except ValueError:
        pass

    # At this point we haven't had any success so lets raise an error
    err = (
        "Packed form designation '{}' could not be unpacked.\n"
        "It could not be recognized as any of the following:\n"
        " - a numbered object (e.g. '03202', 'K3289', '~AZaz')\n"
        " - a provisional designation (e.g. 'J98SC7V', 'K08Aa0A')\n"
        " - a survey designation (e.g. 'PLS2040', 'T1S3138')"
    )
    raise ValueError(err.format(designation_pf))



import numpy as np
import pytest
from scipy.spatial.distance import mahalanobis

from ..cartesian import CartesianCoordinates, CoordinateCovariances
from ..origin import Origin
from ..residuals import (
    Residuals,
    _batch_coords_and_covariances,
    apply_cosine_latitude_correction,
    bound_longitude_residuals,
    calculate_chi2,
    calculate_reduced_chi2,
)


def test_calculate_chi2():
    # Einsums are tricky to follow, so here's a test case
    # with a small number of dimensions and residuals to check that
    # the chi2 calculation is correct.
    residuals = np.array(
        [
            [1, 2, 3],
            [2, 1, 1],
        ]
    )
    covariances = np.array(
        [
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ]
    )
    np.testing.assert_allclose(calculate_chi2(residuals, covariances), [14, 6])

    # Test that the chi2 calculation works for a single residual and is
    # exactly equal to chi2 (observed - expected)^2 / sigma^2
    residuals = np.array([[3]])
    covariances = np.array([[[1]]])
    np.testing.assert_allclose(calculate_chi2(residuals, covariances), [9])


def test_calculate_chi2_missing_diagonal_covariance_values():
    # Lets make sure we raise an error if the covariance matrix has NaNs on the diagonal
    residuals = np.array(
        [
            [1, 2, 3],
            [2, 1, 1],
        ]
    )
    covariances = np.array(
        [
            [[np.nan, 0, 0], [0, np.nan, 0], [0, 0, 1]],
            [[np.nan, 0, 0], [0, np.nan, 0], [0, 0, 1]],
        ]
    )

    with pytest.raises(
        ValueError, match=r"Covariance matrix has NaNs on the diagonal."
    ):
        calculate_chi2(residuals, covariances)


def test_calculate_chi2_missing_off_diagonal_covariance_values():
    # Lets make sure we raise an error if the covariance matrix has NaNs on the off-diagonal
    residuals = np.array(
        [
            [1, 2, 3],
            [2, 1, 1],
        ]
    )
    covariances = np.array(
        [
            [[1, np.nan, 0], [np.nan, 1, 0], [0, 0, 1]],
            [[1, np.nan, 0], [np.nan, 1, 0], [0, 0, 1]],
        ]
    )

    with pytest.warns(
        UserWarning,
        match=r"Covariance matrix has NaNs on the off-diagonal \(these will be assumed to be 0.0\).",
    ):
        np.testing.assert_allclose(calculate_chi2(residuals, covariances), [14, 6])


def test_calculate_chi2_mahalanobis():
    # Test that the calculate_chi2 is equivalent to the Mahalanobis distance squared
    observed = np.array([[1, 1, 1], [2, 2, 2]])
    predicted = np.array([[0, 0, 0], [0, 0, 0]])

    covariances = np.array(
        [
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ]
    )

    mahalanobis_distance = np.zeros(2)
    for i in range(2):
        mahalanobis_distance[i] = mahalanobis(
            observed[i], predicted[i], np.linalg.inv(covariances[i])
        )

    assert np.allclose(
        calculate_chi2(observed - predicted, covariances), mahalanobis_distance**2
    )


def test_batch_coords_and_covariances_single_batch_no_missing_values():
    # Single batch, all dimensions have values
    coords = np.array(
        [
            [1.0, 2.0, 3.0],
            [2.0, 3.0, 4.0],
        ]
    )
    covariances = np.array(
        [
            [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
            [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
        ]
    )

    (
        batch_indices,
        batch_dimensions,
        batch_coords,
        batch_covariances,
    ) = _batch_coords_and_covariances(coords, covariances)
    assert (
        len(batch_indices)
        == len(batch_dimensions)
        == len(batch_coords)
        == len(batch_covariances)
        == 1
    )
    np.testing.assert_equal(batch_indices[0], np.array([0, 1]))
    np.testing.assert_equal(batch_dimensions[0], np.array([0, 1, 2]))
    np.testing.assert_equal(batch_coords[0], coords)
    np.testing.assert_equal(batch_covariances[0], covariances)


def test_batch_coords_and_covariances_single_batch_missing_values():
    # Single batch, one dimension has no values
    coords = np.array(
        [
            [1.0, np.NaN, 3.0],
            [2.0, np.NaN, 4.0],
        ]
    )
    covariances = np.array(
        [
            [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
            [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
        ]
    )

    (
        batch_indices,
        batch_dimensions,
        batch_coords,
        batch_covariances,
    ) = _batch_coords_and_covariances(coords, covariances)
    assert (
        len(batch_indices)
        == len(batch_dimensions)
        == len(batch_coords)
        == len(batch_covariances)
        == 1
    )
    np.testing.assert_equal(batch_indices[0], np.array([0, 1]))
    np.testing.assert_equal(batch_dimensions[0], np.array([0, 2]))
    np.testing.assert_equal(batch_coords[0], np.array([[1.0, 3.0], [2.0, 4.0]]))
    np.testing.assert_equal(
        batch_covariances[0],
        np.array([[[1.0, 0.0], [0.0, 3.0]], [[2.0, 0.0], [0.0, 4.0]]]),
    )

    # Single batch, two dimensions have no values
    coords = np.array(
        [
            [np.NaN, np.NaN, 3.0],
            [np.NaN, np.NaN, 4.0],
        ]
    )
    covariances = np.array(
        [
            [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
            [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
        ]
    )

    (
        batch_indices,
        batch_dimensions,
        batch_coords,
        batch_covariances,
    ) = _batch_coords_and_covariances(coords, covariances)
    assert (
        len(batch_indices)
        == len(batch_dimensions)
        == len(batch_coords)
        == len(batch_covariances)
        == 1
    )
    np.testing.assert_equal(batch_indices[0], np.array([0, 1]))
    np.testing.assert_equal(batch_dimensions[0], np.array([2]))
    np.testing.assert_equal(batch_coords[0], np.array([[3.0], [4.0]]))
    np.testing.assert_equal(batch_covariances[0], np.array([[[3.0]], [[4.0]]]))


def test_batch_coords_and_covariances_multiple_batches():
    # Multiple batches, different rows have different missing values
    coords = np.array(
        [
            [1.0, np.NaN, 3.0],
            [np.NaN, 3.0, 4.0],
        ]
    )
    covariances = np.array(
        [
            [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]],
            [[2.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 4.0]],
        ]
    )

    (
        batch_indices,
        batch_dimensions,
        batch_coords,
        batch_covariances,
    ) = _batch_coords_and_covariances(coords, covariances)
    assert (
        len(batch_indices)
        == len(batch_dimensions)
        == len(batch_coords)
        == len(batch_covariances)
        == 2
    )
    np.testing.assert_equal(batch_indices[0], np.array([0]))
    np.testing.assert_equal(batch_dimensions[0], np.array([0, 2]))
    np.testing.assert_equal(batch_coords[0], np.array([[1.0, 3.0]]))
    np.testing.assert_equal(batch_covariances[0], np.array([[[1.0, 0.0], [0.0, 3.0]]]))

    np.testing.assert_equal(batch_indices[1], np.array([1]))
    np.testing.assert_equal(batch_dimensions[1], np.array([1, 2]))
    np.testing.assert_equal(batch_coords[1], np.array([[3.0, 4.0]]))
    np.testing.assert_equal(batch_covariances[1], np.array([[[3.0, 0.0], [0.0, 4.0]]]))


@pytest.fixture
def observed_array():
    observed_array = np.array(
        [
            [0.2, np.nan, np.nan, np.nan, np.nan, np.nan],
            [0.6, 1.0, 2.0, np.nan, np.nan, np.nan],
            [np.nan, 3.0, np.nan, 4.0, np.nan, np.nan],
            [0.5, 3.0, 0.5, 4.5, 0.1, 0.1],
        ]
    )
    return observed_array


@pytest.fixture
def predicted_array():
    predicted_array = np.array(
        [
            [0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
            [0.5, 1.1, 1.9, 0.2, 0.1, 0.1],
            [0.5, 2.9, 0.2, 4.1, 0.1, 0.1],
            [0.5, 3.0, 0.5, 4.5, 0.1, 0.1],
        ]
    )
    return predicted_array


@pytest.fixture
def expected_residuals(observed_array, predicted_array):
    return observed_array - predicted_array


def test_Residuals_calculate(observed_array, predicted_array, expected_residuals):
    # Test that Residuals.calculate correctly identifies the number of degrees of freedom,
    # and correctly identifies the dimensions that have valid values and those that do not.
    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        covariance=CoordinateCovariances.from_sigmas(np.full((4, 6), 0.1)),
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )

    residuals = Residuals.calculate(observed, predicted)

    assert len(residuals) == 4
    assert residuals.to_array().shape == (4, 6)
    np.testing.assert_equal(residuals.to_array(), expected_residuals)
    assert residuals.dof.to_pylist() == [1, 3, 2, 6]
    np.testing.assert_almost_equal(
        residuals.chi2.to_numpy(zero_copy_only=False), np.array([1, 3, 2, 0])
    )

    # Test that the probabilities for the first and last case are correct (these are more well known examples)
    actual_probabilities = residuals.probability.to_numpy(zero_copy_only=False)
    np.testing.assert_almost_equal(actual_probabilities[0], 0.31731050786291415)
    np.testing.assert_almost_equal(actual_probabilities[3], 1.0)


def test_Residuals_calculate_missing_covariance_values(
    observed_array, predicted_array, expected_residuals
):
    # Test that Residuals.calculate correctly identifies the number of degrees of freedom,
    # and correctly identifies the dimensions that have valid values and those that do not.
    # Here all covariance values (both variates and covariates) are defined
    # for those dimensions that have values, the rest are NaN
    observed_covariances = np.array(
        [
            [
                [0.01, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [0.01, 0.0, 0.0, np.nan, np.nan, np.nan],
                [0.0, 0.01, 0.0, np.nan, np.nan, np.nan],
                [0.0, 0.0, 0.01, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, 0.01, np.nan, 0.0, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, 0.0, np.nan, 0.01, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [0.01, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.01, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.01, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.01, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.01, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.01],
            ],
        ]
    )

    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        covariance=CoordinateCovariances.from_matrix(observed_covariances),
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )

    residuals = Residuals.calculate(observed, predicted)

    assert len(residuals) == 4
    assert residuals.to_array().shape == (4, 6)
    np.testing.assert_equal(residuals.to_array(), expected_residuals)
    assert residuals.dof.to_pylist() == [1, 3, 2, 6]
    np.testing.assert_almost_equal(
        residuals.chi2.to_numpy(zero_copy_only=False), np.array([1, 3, 2, 0])
    )

    # Test that the probabilities for the first and last case are correct (these are more well known examples)
    actual_probabilities = residuals.probability.to_numpy(zero_copy_only=False)
    np.testing.assert_almost_equal(actual_probabilities[0], 0.31731050786291415)
    np.testing.assert_almost_equal(actual_probabilities[3], 1.0)


def test_Residuals_calculate_missing_off_diagonal_covariance_values(
    observed_array, predicted_array, expected_residuals
):
    # Test that Residuals.calculate correctly identifies the number of degrees of freedom,
    # and correctly identifies the dimensions that have valid values and those that do not.
    # Here only variance values are defined
    # for those dimensions that have values, the covariates and the rest are NaN
    observed_covariances = np.array(
        [
            [
                [0.01, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [0.01, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, 0.01, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, 0.01, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, 0.01, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, 0.01, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
            [
                [0.01, np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, 0.01, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, 0.01, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, 0.01, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, 0.01, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan, 0.01],
            ],
        ]
    )

    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        covariance=CoordinateCovariances.from_matrix(observed_covariances),
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )

    with pytest.warns(
        UserWarning,
        match=r"Covariance matrix has NaNs on the off-diagonal \(these will be assumed to be 0.0\).",
    ):
        residuals = Residuals.calculate(observed, predicted)

    assert len(residuals) == 4
    assert residuals.to_array().shape == (4, 6)
    np.testing.assert_equal(residuals.to_array(), expected_residuals)
    assert residuals.dof.to_pylist() == [1, 3, 2, 6]
    np.testing.assert_almost_equal(
        residuals.chi2.to_numpy(zero_copy_only=False), np.array([1, 3, 2, 0])
    )

    # Test that the probabilities for the first and last case are correct (these are more well known examples)
    actual_probabilities = residuals.probability.to_numpy(zero_copy_only=False)
    np.testing.assert_almost_equal(actual_probabilities[0], 0.31731050786291415)
    np.testing.assert_almost_equal(actual_probabilities[3], 1.0)


def test_Residuals_calculate_raises_frames():
    # Test that an error is raised when the frames are not equal
    observed_array = np.random.random((10, 6))
    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(10, "SUN", dtype="object")),
        frame="ecliptic",
    )

    predicted_array = np.random.random((10, 6))
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(10, "SUN", dtype="object")),
        frame="equatorial",
    )

    with pytest.raises(ValueError, match=r"coordinates must have the same frame."):
        Residuals.calculate(observed, predicted)


def test_Residuals_calculate_raises_origins():
    # Test that an error is raised when the frames are not equal
    observed_array = np.random.random((10, 6))
    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(10, "SUN", dtype="object")),
        frame="equatorial",
    )

    predicted_array = np.random.random((10, 6))
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(10, "EARTH", dtype="object")),
        frame="equatorial",
    )

    with pytest.raises(ValueError, match=r"coordinates must have the same origin."):
        Residuals.calculate(observed, predicted)


def test_apply_cosine_latitude_correction():
    # Test that apply_cosine_latitude_correction correctly applies the cosine latitude correction
    # to the residuals and covariances as a function of the latitude.
    residual_array = np.array(
        [
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
        ],
        dtype=np.float64,
    )

    # Create covariances that are all 1
    covariance_array = np.ones((4, 6, 6), dtype=np.float64)

    # Cosine of 0 degrees is 1
    # Cosine of 45 degrees is 1/sqrt(2)
    # Cosien of 60 degrees is 1/2
    # Cosine of 90 degrees is 0
    # The latter represents an unphysical case
    latitude_array = np.array([0, 45, 60, 90])
    cos_latitude = np.cos(np.radians(latitude_array))
    expected_residual_array = np.array(
        [
            [0, 10, 0, 0, 1, 0],
            [0, 10 / np.sqrt(2), 0, 0, 1 / np.sqrt(2), 0],
            [0, 5, 0, 0, 1 / 2, 0],
            [0, 0, 0, 0, 0, 0],
        ],
        dtype=np.float64,
    )

    # Covariance matrix elements
    # rho_rho  (:,0,0), rho_lon  (: 0,1), rho_lat  (:,0,2), rho_vrho  (:,0,3), rho_vlon  (:,0,4), rho_vlat  (:,0,5)
    # lon_rho  (:,1,0), lon_lon  (:,1,1), lon_lat  (:,1,2), lon_vrho  (:,1,3), lon_vlon  (:,1,4), lon_vlat  (:,1,5)
    # lat_rho  (:,2,0), lat_lon  (:,2,1), lat_lat  (:,2,2), lat_vrho  (:,2,3), lat_vlon  (:,2,4), lat_vlat  (:,2,5)
    # vrho_rho (:,3,0), vrho_lon (:,3,1), vrho_lat (:,3,2), vrho_vrho (:,3,3), vrho_vlon (:,3,4), vrho_vlat (:,3,5)
    # vlon_rho (:,4,0), vlon_lon (:,4,1), vlon_lat (:,4,2), vlon_vrho (:,4,3), vlon_vlon (:,4,4), vlon_vlat (:,4,5)
    # vlat_rho (:,5,0), vlat_lon (:,5,1), vlat_lat (:,5,2), vlat_vrho (:,5,3), vlat_vlon (:,5,4), vlat_vlat (:,5,5)
    # The only elements that should change are those rows and columns containing longitude and longitudinal velocity
    # Not that for cov_lon_lon and cov_vlon_vlon the correction is applied twice as expected (we touch both quantities
    # as rows and columns)
    expected_covariance = np.ones_like(covariance_array)
    expected_covariance[:, :, 1] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, :, 4] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, 1, :] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, 4, :] *= cos_latitude[:, np.newaxis]

    corrected_residuals, corrected_covariances = apply_cosine_latitude_correction(
        latitude_array, residual_array, covariance_array
    )

    # Test that the residuals are corrected correctly
    np.testing.assert_almost_equal(
        corrected_residuals, expected_residual_array, decimal=15
    )

    # Test that the covariances are corrected correctly
    np.testing.assert_almost_equal(
        corrected_covariances, expected_covariance, decimal=15
    )

    # Add additional assertions for clarity
    # The above are sufficient to test the code but obfuscate the linear algebra

    # We expect that the first covariance matrix is completely unchanged
    np.testing.assert_equal(corrected_covariances[:, 0, 0], covariance_array[:, 0, 0])

    # Define the expected diagonal and off-diagonal elements that would have changed
    expected_covariates = np.array([1, 1 / np.sqrt(2), 1 / 2, 0])
    expected_variates = expected_covariates**2

    # Test_cov_lon_rho and rho_lon (1, 2)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 0], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 1], expected_covariates, decimal=15
    )
    # Test cov_lon_lon (3)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 1], expected_variates, decimal=15
    )
    # Test cov_lon_lat and lat_lon (4, 5)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 2], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 1], expected_covariates, decimal=15
    )
    # Test cov_lon_vrho and vrho_lon (6, 7)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 3], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 1], expected_covariates, decimal=15
    )
    # Test cov_lon_vlon and vlon_lon (8, 9) - note that this is applied twice
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 4], expected_variates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 1], expected_variates, decimal=15
    )
    # Test cov_lon_vlat and vlat_lon (10, 11)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 1, 5], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 1], expected_covariates, decimal=15
    )

    # Test cov_vlon_rho and rho_vlon (12, 13)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 0], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 4], expected_covariates, decimal=15
    )
    # Test cov_vlon_lat and lat_vlon (14, 15)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 2], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 4], expected_covariates, decimal=15
    )
    # Test cov_vlon_vlon (16)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 4], expected_variates, decimal=15
    )
    # Test cov vlon_vrho and vrho_vlon (17, 18)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 3], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 4], expected_covariates, decimal=15
    )
    # Test cov_vlon_vlat and vlat_vlon (19, 20)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 4, 5], expected_covariates, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 4], expected_covariates, decimal=15
    )

    # Define the expected diagonal and off-diagonal elements that would not have changed
    expected_covariates_unchanged = expected_variates_unchanged = np.ones(4)
    # Test cov_rho_rho (21)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 0], expected_variates_unchanged, decimal=15
    )
    # Test cov_rho_lat and lat_rho (22, 23)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 2], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 0], expected_covariates_unchanged, decimal=15
    )
    # Test cov_rho_vrho and vrho_rho (24, 25)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 3], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 0], expected_covariates_unchanged, decimal=15
    )
    # Test cov_rho_vlat and vlat_rho (26, 27)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 0, 5], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 0], expected_covariates_unchanged, decimal=15
    )
    # Test cov_lat_lat (28)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 2], expected_variates_unchanged, decimal=15
    )
    # Test cov_lat_vrho and vrho_lat (29, 30)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 3], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 2], expected_covariates_unchanged, decimal=15
    )
    # Test cov_lat_vlat and vlat_lat (31, 32)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 2, 5], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 2], expected_covariates_unchanged, decimal=15
    )
    # Test cov_vrho_vrho (33)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 3], expected_variates_unchanged, decimal=15
    )
    # Test cov_vrho_vlat and vlat_vrho (34, 35)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 3, 5], expected_covariates_unchanged, decimal=15
    )
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 3], expected_covariates_unchanged, decimal=15
    )
    # Test cov_vlat_vlat (36)
    np.testing.assert_almost_equal(
        corrected_covariances[:, 5, 5], expected_variates_unchanged, decimal=15
    )


def test_apply_cosine_latitude_correction_missing_off_diagonal_covariance_values():
    # Test that apply_cosine_latitude_correction correctly applies the cosine latitude correction
    # to the residuals and covariances as a function of the latitude. Specifically, lets make sure
    # that if the covariance matrix has NaNs on the off-diagonal that these are correctly handled
    residual_array = np.array(
        [
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
            [0, 10, 0, 0, 1, 0],
        ],
        dtype=np.float64,
    )

    # Create covariances that only have the diagonal defined
    covariance_array = np.full((4, 6, 6), np.nan, dtype=np.float64)
    covariance_array[:, np.arange(6), np.arange(6)] = 1.0

    # Cosine of 0 degrees is 1
    # Cosine of 45 degrees is 1/sqrt(2)
    # Cosien of 60 degrees is 1/2
    # Cosine of 90 degrees is 0
    # The latter represents an unphysical case
    latitude_array = np.array([0, 45, 60, 90])
    cos_latitude = np.cos(np.radians(latitude_array))
    expected_residual_array = np.array(
        [
            [0, 10, 0, 0, 1, 0],
            [0, 10 / np.sqrt(2), 0, 0, 1 / np.sqrt(2), 0],
            [0, 5, 0, 0, 1 / 2, 0],
            [0, 0, 0, 0, 0, 0],
        ],
        dtype=np.float64,
    )

    # Covariance matrix elements
    # rho_rho  (:,0,0), rho_lon  (: 0,1), rho_lat  (:,0,2), rho_vrho  (:,0,3), rho_vlon  (:,0,4), rho_vlat  (:,0,5)
    # lon_rho  (:,1,0), lon_lon  (:,1,1), lon_lat  (:,1,2), lon_vrho  (:,1,3), lon_vlon  (:,1,4), lon_vlat  (:,1,5)
    # lat_rho  (:,2,0), lat_lon  (:,2,1), lat_lat  (:,2,2), lat_vrho  (:,2,3), lat_vlon  (:,2,4), lat_vlat  (:,2,5)
    # vrho_rho (:,3,0), vrho_lon (:,3,1), vrho_lat (:,3,2), vrho_vrho (:,3,3), vrho_vlon (:,3,4), vrho_vlat (:,3,5)
    # vlon_rho (:,4,0), vlon_lon (:,4,1), vlon_lat (:,4,2), vlon_vrho (:,4,3), vlon_vlon (:,4,4), vlon_vlat (:,4,5)
    # vlat_rho (:,5,0), vlat_lon (:,5,1), vlat_lat (:,5,2), vlat_vrho (:,5,3), vlat_vlon (:,5,4), vlat_vlat (:,5,5)
    # The only elements that should change are those rows and columns containing longitude and longitudinal velocity
    # Not that for cov_lon_lon and cov_vlon_vlon the correction is applied twice as expected (we touch both quantities
    # as rows and columns)
    expected_covariance = covariance_array.copy()
    expected_covariance[:, :, 1] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, :, 4] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, 1, :] *= cos_latitude[:, np.newaxis]
    expected_covariance[:, 4, :] *= cos_latitude[:, np.newaxis]

    corrected_residuals, corrected_covariances = apply_cosine_latitude_correction(
        latitude_array, residual_array, covariance_array
    )

    # Test that the residuals are corrected correctly
    np.testing.assert_almost_equal(
        corrected_residuals, expected_residual_array, decimal=15
    )

    # Test that the covariances are corrected correctly
    np.testing.assert_almost_equal(
        corrected_covariances, expected_covariance, decimal=15
    )


def test_bound_longitude_residuals():
    # Test that bound_longitude_residuals correctly bounds the longitude residuals
    # to within -180 and 180 degrees and that it handles the signs correctly across the
    # 0/360 degree boundary. Typically residuals that are positive mean that the observed
    # value is greater than the predicted value, and vice versa for negative residuals.

    # Observed  Predicted  Arithmetic  Expected
    #    1        359        -358         -2
    #  359          1         358          2
    #   60        240        -180       -180
    #  240         60         180        180
    #   10        190        -180       -180
    #  190         10         180        180
    #   60        250        -190       -170
    #  250         60         190        170
    #   60        230        -170       -170
    #  230         60         170        170
    observed_array = np.ones((10, 6))
    observed_array[:, 1] = np.array([1, 359, 60, 240, 10, 190, 60, 250, 60, 230])

    predicted_array = np.ones((10, 6))
    predicted_array[:, 1] = np.array([359, 1, 240, 60, 190, 10, 250, 60, 230, 60])

    residuals_array = observed_array - predicted_array
    residuals = bound_longitude_residuals(observed_array, residuals_array)

    np.testing.assert_equal(
        residuals[:, 1], np.array([-2, 2, -180, 180, -180, 180, -170, 170, -170, 170])
    )


def test_calculate_reduced_chi2(observed_array, predicted_array):
    # Test that Residuals.calculate correctly identifies the number of degrees of freedom,
    # and correctly identifies the dimensions that have valid values and those that do not.
    observed = CartesianCoordinates.from_kwargs(
        x=observed_array[:, 0],
        y=observed_array[:, 1],
        z=observed_array[:, 2],
        vx=observed_array[:, 3],
        vy=observed_array[:, 4],
        vz=observed_array[:, 5],
        covariance=CoordinateCovariances.from_sigmas(np.full((4, 6), 0.1)),
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )
    predicted = CartesianCoordinates.from_kwargs(
        x=predicted_array[:, 0],
        y=predicted_array[:, 1],
        z=predicted_array[:, 2],
        vx=predicted_array[:, 3],
        vy=predicted_array[:, 4],
        vz=predicted_array[:, 5],
        origin=Origin.from_kwargs(code=np.full(4, "SUN", dtype="object")),
        frame="ecliptic",
    )

    residuals = Residuals.calculate(observed, predicted)

    assert residuals.dof.to_pylist() == [1, 3, 2, 6]
    np.testing.assert_almost_equal(
        residuals.chi2.to_numpy(zero_copy_only=False), np.array([1, 3, 2, 0])
    )

    reduced_chi2 = calculate_reduced_chi2(residuals, 6)
    np.testing.assert_almost_equal(reduced_chi2, 6 / 6)

    reduced_chi2 = calculate_reduced_chi2(residuals, 4)
    np.testing.assert_almost_equal(reduced_chi2, 6 / 8)



from importlib.resources import files

import numpy as np
import pandas as pd
import pytest

from ...coordinates.cometary import CometaryCoordinates
from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import Origin
from ...time.time import Timestamp
from ..classification import calc_orbit_class
from ..orbits import Orbits

dynamical_classes_dir = files("adam_core.orbits.tests.testdata")


def sbdb_df_to_cometary(df: pd.DataFrame) -> CometaryCoordinates:
    return CometaryCoordinates.from_kwargs(
        q=df["q"],
        e=df["e"],
        i=df["i"],
        raan=df["om"],
        ap=df["w"],
        tp=df["tp"] - 2400000.5,
        time=Timestamp.from_mjd(df["epoch_mjd"].values, scale="tdb"),
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(df))]),
        frame="ecliptic",
    )


def sbdb_df_to_keplerian(df: pd.DataFrame) -> KeplerianCoordinates:
    return KeplerianCoordinates.from_kwargs(
        a=df["a"],
        e=df["e"],
        i=df["i"],
        raan=df["om"],
        ap=df["w"],
        M=df["ma"],
        time=Timestamp.from_mjd(df["epoch_mjd"].values, scale="tdb"),
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(df))]),
        frame="ecliptic",
    )


def sbdb_df_to_orbits(df: pd.DataFrame) -> Orbits:
    return Orbits.from_kwargs(
        object_id=df["full_name"],
        coordinates=sbdb_df_to_cometary(df).to_cartesian(),
    )


def ieo_sample():
    df = pd.read_csv(dynamical_classes_dir / "ieo.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("IEO", len(df))


def ate_sample():
    df = pd.read_csv(dynamical_classes_dir / "ate.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("ATE", len(df))


def apo_sample():
    df = pd.read_csv(dynamical_classes_dir / "apo.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("APO", len(df))


def amo_sample():
    df = pd.read_csv(dynamical_classes_dir / "amo.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("AMO", len(df))


def mca_sample():
    df = pd.read_csv(dynamical_classes_dir / "mca.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("MCA", len(df))


def imb_sample():
    df = pd.read_csv(dynamical_classes_dir / "imb.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("IMB", len(df))


def mba_sample():
    df = pd.read_csv(dynamical_classes_dir / "mba.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("MBA", len(df))


def omb_sample():
    df = pd.read_csv(dynamical_classes_dir / "omb.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("OMB", len(df))


def tjn_sample():
    df = pd.read_csv(dynamical_classes_dir / "tjn.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("TJN", len(df))


def cen_sample():
    df = pd.read_csv(dynamical_classes_dir / "cen.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("CEN", len(df))


def tno_sample():
    df = pd.read_csv(dynamical_classes_dir / "tno.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("TNO", len(df))


def hya_sample():
    df = pd.read_csv(dynamical_classes_dir / "hya.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("HYA", len(df))


def ast_sample():
    df = pd.read_csv(dynamical_classes_dir / "ast.csv")
    cometary = sbdb_df_to_cometary(df)
    keplerian = sbdb_df_to_keplerian(df)
    orbits = sbdb_df_to_orbits(df)
    return orbits, cometary, keplerian, np.repeat("AST", len(df))


SAMPLES = [
    ieo_sample(),
    ate_sample(),
    apo_sample(),
    amo_sample(),
    mca_sample(),
    imb_sample(),
    mba_sample(),
    omb_sample(),
    tjn_sample(),
    cen_sample(),
    tno_sample(),
    hya_sample(),
    ast_sample(),
]


@pytest.mark.parametrize(
    "sample",
    SAMPLES,
)
def test_calc_orbit_class_cometary(sample):
    _, cometary, _, expected_classes = sample
    classes = calc_orbit_class(cometary)
    np.testing.assert_array_equal(classes, expected_classes)


@pytest.mark.parametrize(
    "sample",
    SAMPLES,
)
def test_calc_orbit_class_keplerian(sample):
    _, _, keplerian, expected_classes = sample
    classes = calc_orbit_class(keplerian)
    np.testing.assert_array_equal(classes, expected_classes)


@pytest.mark.parametrize(
    "sample",
    SAMPLES,
)
def test_calc_orbit_class_orbits(sample):
    orbits, _, _, expected_classes = sample
    classes = orbits.dynamical_class()
    np.testing.assert_array_equal(classes, expected_classes)



import uuid
from typing import Literal, Optional

import numpy as np
import pyarrow.compute as pc
import quivr as qv

from ..coordinates.cartesian import CartesianCoordinates
from ..coordinates.covariances import CoordinateCovariances, weighted_covariance
from ..coordinates.spherical import SphericalCoordinates
from ..coordinates.variants import VariantCoordinatesTable, create_coordinate_variants
from .ephemeris import Ephemeris
from .orbits import Orbits


class VariantOrbits(qv.Table):
    orbit_id = qv.LargeStringColumn(default=lambda: uuid.uuid4().hex)
    object_id = qv.LargeStringColumn(nullable=True)
    variant_id = qv.LargeStringColumn(nullable=True)
    weights = qv.Float64Column(nullable=True)
    weights_cov = qv.Float64Column(nullable=True)
    coordinates = CartesianCoordinates.as_column()

    @classmethod
    def create(
        cls,
        orbits: Orbits,
        method: Literal["auto", "sigma-point", "monte-carlo"] = "auto",
        num_samples: int = 10000,
        alpha: float = 1,
        beta: float = 0,
        kappa: float = 0,
        seed: Optional[int] = None,
    ) -> "VariantOrbits":
        """
        Sample and create variants for the given orbits by sampling the covariance matrices.
        There are three supported methods:
        - sigma-point: Sample the covariance matrix using sigma points. This is the fastest method,
        but can be inaccurate if the covariance matrix is not well behaved.
        - monte-carlo: Sample the covariance matrix using a monte carlo method.
        This is the slowest method, but is the most accurate. 10k samples are drawn.
        - auto: Automatically select the best method based on the covariance matrix.
        If the covariance matrix is well behaved then sigma-point sampling will be used.
        If the covariance matrix is not well behaved then monte-carlo sampling will be used.

        When sampling with monte-carlo, 10k samples are drawn. Sigma-point sampling draws 13 samples
        for 6-dimensional coordinates.

        Parameters
        ----------
        orbits : '~adam_core.orbits.orbits.Orbits'
            The orbits for which to create variant orbits.
        method : {'sigma-point', 'monte-carlo', 'auto'}, optional
            The method to use for sampling the covariance matrix. If 'auto' is selected then the method
            will be automatically selected based on the covariance matrix. The default is 'auto'.
        num_samples : int, optional
            The number of samples to draw when sampling with monte-carlo.
        alpha : float, optional
            Spread of the sigma points between 1e^-2 and 1.
        beta : float, optional
            Prior knowledge of the distribution when generating sigma points usually set to 2 for a Gaussian.
        kappa : float, optional
            Secondary scaling parameter when generating sigma points usually set to 0.

        Returns
        -------
        variants_orbits : '~adam_core.orbits.variants.VariantOrbits'
            The variant orbits.
        """
        variant_coordinates: VariantCoordinatesTable = create_coordinate_variants(
            orbits.coordinates,
            method=method,
            num_samples=num_samples,
            alpha=alpha,
            beta=beta,
            kappa=kappa,
            seed=seed,
        )

        return cls.from_kwargs(
            orbit_id=pc.take(orbits.orbit_id, variant_coordinates.index),
            object_id=pc.take(orbits.object_id, variant_coordinates.index),
            variant_id=np.array(
                np.arange(len(variant_coordinates)).astype(str), dtype="object"
            ),
            weights=variant_coordinates.weight,
            weights_cov=variant_coordinates.weight_cov,
            coordinates=variant_coordinates.sample,
        )

    def link_to_orbits(
        self, orbits: Orbits
    ) -> qv.MultiKeyLinkage[Orbits, "VariantOrbits"]:
        """
        Link variants to the orbits from which they were generated.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits`
            Orbits from which the variants were generated.

        Returns
        -------
        linkage : `~quivr.MultiKeyLinkage[Orbits, VariantOrbits]`
            Linkage between variants and orbits.
        """
        assert orbits.coordinates.time.scale == self.coordinates.time.scale

        # We might want to replace linking on jd1 and jd2 with just linking on mjd
        # once the changes have been merged
        return qv.MultiKeyLinkage(
            orbits,
            self,
            left_keys={
                "orbit_id": orbits.orbit_id,
                "day": orbits.coordinates.time.days,
                "millis": orbits.coordinates.time.millis(),
            },
            right_keys={
                "orbit_id": self.orbit_id,
                "day": self.coordinates.time.days,
                "millis": self.coordinates.time.millis(),
            },
        )

    def collapse(self, orbits: Orbits) -> Orbits:
        """
        Collapse the variants and recalculate the covariance matrix for each
        each orbit at each epoch. The mean state is taken from the orbits class and
        is not calculated from the variants.

        Parameters
        ----------
        orbits : `~adam_core.orbits.orbits.Orbits`
            Orbits from which the variants were generated.

        Returns
        -------
        collapsed_orbits : `~adam_core.orbits.orbits.Orbits`
            The collapsed orbits.
        """
        link = self.link_to_orbits(orbits)

        # Iterate over the variants and calculate the mean state and covariance matrix
        # for each orbit at each epoch then create a new orbit with the calculated covariance matrix
        orbits_list = []
        for orbit in orbits:
            assert len(orbit) == 1

            key = link.key(
                orbit_id=orbit.orbit_id[0].as_py(),
                day=orbit.coordinates.time.days[0].as_py(),
                millis=orbit.coordinates.time.millis()[0].as_py(),
            )
            variants = link.select_right(key)

            samples = variants.coordinates.values
            mean = orbit.coordinates.values[0]
            covariance = weighted_covariance(
                mean, samples, variants.weights_cov.to_numpy(zero_copy_only=False)
            ).reshape(1, 6, 6)

            orbit_collapsed = orbit.set_column(
                "coordinates.covariance", CoordinateCovariances.from_matrix(covariance)
            )

            orbits_list.append(orbit_collapsed)

        return qv.concatenate(orbits_list)


class VariantEphemeris(qv.Table):

    orbit_id = qv.LargeStringColumn(default=lambda: uuid.uuid4().hex)
    object_id = qv.LargeStringColumn(nullable=True)
    variant_id = qv.LargeStringColumn(nullable=True)
    weights = qv.Float64Column(nullable=True)
    weights_cov = qv.Float64Column(nullable=True)
    coordinates = SphericalCoordinates.as_column()

    def link_to_ephemeris(
        self, ephemeris: Ephemeris
    ) -> qv.MultiKeyLinkage[Ephemeris, "VariantEphemeris"]:
        """
        Link variants to the ephemeris for which they were generated.

        Parameters
        ----------
        ephemeris : `~adam_core.orbits.ephemeris.Ephemeris`
            Ephemeris for which the variants were generated.

        Returns
        -------
        linkage : `~quivr.MultiKeyLinkage[Ephemeris, EphemerisVariants]`
            Linkage between variants and ephemeris.
        """
        assert ephemeris.coordinates.time.scale == self.coordinates.time.scale

        # We might want to replace linking on jd1 and jd2 with just linking on mjd
        # once the changes have been merged
        return qv.MultiKeyLinkage(
            ephemeris,
            self,
            left_keys={
                "orbit_id": ephemeris.orbit_id,
                "day": ephemeris.coordinates.time.days,
                "millis": ephemeris.coordinates.time.millis(),
            },
            right_keys={
                "orbit_id": self.orbit_id,
                "day": self.coordinates.time.days,
                "millis": self.coordinates.time.millis(),
            },
        )

    def collapse(self, ephemeris: Ephemeris) -> Ephemeris:
        """
        Collapse the variants and recalculate the covariance matrix for each
        each ephemeris at each epoch. The mean state is taken from the ephemeris class and
        is not calculate from the variants.

        Parameters
        ----------
        ephemeris : `~adam_core.orbits.ephemeris.Ephemeris`
            Ephemeris for which the variants were generated.

        Returns
        -------
        collapsed_ephemeris : `~adam_core.orbits.ephemeris.Ephemeris`
            The collapsed ephemeris (with covariance matrices calculated based
            on the samples).
        """
        link = self.link_to_ephemeris(ephemeris)

        # Iterate over the variants and calculate the mean state and covariance matrix
        # for each orbit at each epoch then create a new orbit with the calculated covariance matrix
        ephemeris_list = []
        for ephemeris_i in ephemeris:
            assert len(ephemeris_i) == 1

            key = link.key(
                orbit_id=ephemeris_i.orbit_id[0].as_py(),
                day=ephemeris_i.coordinates.time.days[0].as_py(),
                millis=ephemeris_i.coordinates.time.millis()[0].as_py(),
            )
            variants = link.select_right(key)

            samples = variants.coordinates.values
            mean = ephemeris_i.coordinates.values[0]
            covariance = weighted_covariance(
                mean, samples, variants.weights_cov.to_numpy(zero_copy_only=False)
            ).reshape(1, 6, 6)

            ephemeris_collapsed = ephemeris_i.set_column(
                "coordinates.covariance", CoordinateCovariances.from_matrix(covariance)
            )

            ephemeris_list.append(ephemeris_collapsed)

        return qv.concatenate(ephemeris_list)



import logging
from enum import Enum

import numpy as np
import pyarrow.compute as pc
import quivr as qv

from ..constants import KM_P_AU, S_P_DAY

logger = logging.getLogger(__name__)


class OriginCodes(Enum):
    SOLAR_SYSTEM_BARYCENTER = 0
    MERCURY_BARYCENTER = 1
    VENUS_BARYCENTER = 2
    EARTH_MOON_BARYCENTER = 3
    MARS_BARYCENTER = 4
    JUPITER_BARYCENTER = 5
    SATURN_BARYCENTER = 6
    URANUS_BARYCENTER = 7
    NEPTUNE_BARYCENTER = 8
    SUN = 10
    MERCURY = 199
    VENUS = 299
    EARTH = 399
    MOON = 301
    MARS = 499
    JUPITER = 599
    SATURN = 699
    URANUS = 799
    NEPTUNE = 899


def _convert_mu_units(mu: float):
    """
    Convert mu in km^3 / s^2 to au^3 / day^2.
    """
    return mu / KM_P_AU**3 * S_P_DAY**2


class OriginGravitationalParameters(float, Enum):
    # Taken from https://iopscience.iop.org/article/10.3847/1538-3881/abd414
    MERCURY_BARYCENTER = _convert_mu_units(22032.080486418)
    VENUS_BARYCENTER = _convert_mu_units(324858.592000)
    MARS_BARYCENTER = _convert_mu_units(42828.375816)
    JUPITER_BARYCENTER = _convert_mu_units(126712764.100000)
    SATURN_BARYCENTER = _convert_mu_units(37940584.841800)
    URANUS_BARYCENTER = _convert_mu_units(5794556.400000)
    NEPTUNE_BARYCENTER = _convert_mu_units(6836527.100580)
    PLUTO_BARYCENTER = _convert_mu_units(975.500000)
    SUN = _convert_mu_units(132712440041.279419)
    MERCURY = _convert_mu_units(22031.868551)
    VENUS = _convert_mu_units(324858.592000)
    EARTH = _convert_mu_units(398600.435507)
    MOON = _convert_mu_units(4902.800118)

    @classmethod
    def SOLAR_SYSTEM_BARYCENTER(cls) -> float:
        """
        Return the gravitational parameter of the Solar System barycenter as approximated
        by adding the gravitational parameters of the Sun, Mercury, Venus, Earth, Moon,
        Mars, Jupiter, Uranus, and Neptune.

        Returns
        -------
        mu : float
            The gravitational parameter of the Solar System barycenter in au^3 / day^2.
        """
        return (
            cls.SUN
            + cls.MERCURY_BARYCENTER
            + cls.VENUS_BARYCENTER
            + cls.EARTH
            + cls.MOON
            + cls.MARS_BARYCENTER
            + cls.JUPITER_BARYCENTER
            + cls.URANUS_BARYCENTER
            + cls.NEPTUNE_BARYCENTER
            + cls.PLUTO_BARYCENTER
        )


# TODO: Replace with DictionaryColumn or similar
#       Investigate whether this class is even necessary
class Origin(qv.Table):
    code = qv.LargeStringColumn()

    def __eq__(self, other: object) -> np.ndarray:
        if isinstance(other, (str, np.ndarray)):
            codes = self.code.to_numpy(zero_copy_only=False)
            return codes == other
        elif isinstance(other, OriginCodes):
            codes = self.code.to_numpy(zero_copy_only=False)
            return codes == other.name
        elif isinstance(other, Origin):
            codes = self.code.to_numpy(zero_copy_only=False)
            other_codes = other.code.to_numpy(zero_copy_only=False)
            return codes == other_codes
        else:
            raise TypeError(f"Cannot compare Origin to type: {type(other)}")

    def __ne__(self, other: object) -> np.ndarray:
        return ~self.__eq__(other)

    def mu(self) -> np.ndarray:
        """
        Return the gravitational parameter of the origin.

        Returns
        -------
        mu : `~numpy.ndarray` (N)
            The gravitational parameter of the origin in au^3 / day^2.

        Raises
        ------
        ValueError
            If the origin code is not recognized.
        """
        mu = np.empty(len(self.code), dtype=np.float64)
        for code in pc.unique(self.code):
            code = code.as_py()
            mask = pc.equal(self.code, code).to_numpy(zero_copy_only=False)
            if code == "SOLAR_SYSTEM_BARYCENTER":
                mu[mask] = OriginGravitationalParameters.SOLAR_SYSTEM_BARYCENTER()
            elif code in OriginGravitationalParameters.__members__:
                mu[mask] = OriginGravitationalParameters[code].value
            else:
                raise ValueError(f"Unknown origin code: {code}")

        return mu



from importlib.resources import files

import numpy as np
import pandas as pd
import pyarrow.compute as pc

from ..observations import make_observations


def test_make_observations():
    # Test that we get two tables from make observations are
    # the expected length
    exposures, detections, associations = make_observations()
    assert len(exposures) == 2412
    assert len(detections) == 2520
    assert len(associations) == 2520

    # Load original ephemeris file
    ephemeris = pd.read_csv(
        files("adam_core.utils.helpers.data").joinpath("ephemeris.csv")
    )
    ephemeris.sort_values(by=["mjd_utc", "observatory_code"], inplace=True)
    ephemeris["obs_id"] = [f"obs_{i:04d}" for i in range(len(ephemeris))]

    # Test that the RAs, Decs, and associations were correctly set
    for object_id in associations.object_id.unique().to_numpy(zero_copy_only=False):

        # Get the obs_ids for this object from the associations table
        associations_mask = pc.equal(associations.object_id, object_id)
        associations_object = associations.apply_mask(associations_mask)
        obs_ids = associations_object.detection_id

        # Get the RA and Decs for this object from the detections table
        detections_mask = pc.is_in(detections.id, obs_ids)
        detections_object = detections.apply_mask(detections_mask)
        ras = detections_object.ra.to_numpy(zero_copy_only=False)
        decs = detections_object.dec.to_numpy(zero_copy_only=False)

        # Get the input ephemeris for this object
        ephemeris_object = ephemeris[ephemeris["targetname"] == object_id]
        obs_ids_ephemeris = ephemeris_object["obs_id"].values
        ras_ephemeris = ephemeris_object["RA"].values
        decs_ephemeris = ephemeris_object["DEC"].values

        # Test that the obs_ids are identical
        np.testing.assert_array_equal(obs_ids, obs_ids_ephemeris)

        # Test that the RA and Decs are close
        np.testing.assert_allclose(ras, ras_ephemeris, atol=1e-15, rtol=0)
        np.testing.assert_allclose(decs, decs_ephemeris, atol=1e-15, rtol=0)



from typing import Union

import numpy as np

from ..coordinates.cometary import CometaryCoordinates
from ..coordinates.keplerian import KeplerianCoordinates


def calc_orbit_class(
    elements: Union[KeplerianCoordinates, CometaryCoordinates]
) -> np.ndarray:
    """
    Calculate the orbital class for each Keplerian or Cometary orbit.

    Based on the classification scheme defined by the Planetary Data System Small
    Bodies Node (see: https://pdssbn.astro.umd.edu/data_other/objclass.shtml).

    TODO: Classification is currently limited to asteroid dynamical classes. Cometary class
    have not yet been implemented.

    Parameters
    ----------
    elements : KeplerianCoordinates or CometaryCoordinates
        Keplerian orbits for which to find classes.

    Returns
    -------
    orbit_class : `~numpy.ndarray`
        Class for each orbit.
    """
    if isinstance(elements, CometaryCoordinates):
        a = elements.a
        e = elements.e.to_numpy(zero_copy_only=False)
        q = elements.q.to_numpy(zero_copy_only=False)
        Q = elements.Q

    elif isinstance(elements, KeplerianCoordinates):
        a = elements.a.to_numpy(zero_copy_only=False)
        e = elements.e.to_numpy(zero_copy_only=False)
        q = elements.q
        Q = elements.Q

    orbit_class = np.array(["AST" for i in range(len(elements))])

    orbit_class_dict = {
        "AMO": np.where((a > 1.0) & (q > 1.017) & (q < 1.3)),
        "APO": np.where((a > 1.0) & (q < 1.017)),
        "ATE": np.where((a < 1.0) & (Q > 0.983)),
        "CEN": np.where((a > 5.5) & (a < 30.1)),
        "IEO": np.where((Q < 0.983)),
        "IMB": np.where((a < 2.0) & (q > 1.666)),
        "MBA": np.where((a > 2.0) & (a < 3.2) & (q > 1.666)),
        "MCA": np.where((a < 3.2) & (q > 1.3) & (q < 1.666)),
        "OMB": np.where((a > 3.2) & (a < 4.6)),
        "TJN": np.where((a > 4.6) & (a < 5.5) & (e < 0.3)),
        "TNO": np.where((a > 30.1)),
        "PAA": np.where((e == 1)),
        "HYA": np.where((e > 1)),
    }
    for c, v in orbit_class_dict.items():
        orbit_class[v] = c

    return orbit_class
